#!/usr/bin/env python3

#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from utopia2output import *
from time import sleep
from pyb00st.movehub import MoveHub
from pyb00st.constants import *

mymovehub = MoveHub('00:16:53:a5:6f:8a', 'Auto', 'hci0')

try:
   mymovehub.start()
   mymovehub.subscribe_all()
   mymovehub.listen_angle_sensor(PORT_C)
   mymovehub.listen_angle_sensor(PORT_A)
   mymovehub.listen_angle_sensor(PORT_B)

except:
    print('Cannot connect to Boost.')


# define our action functions
def head (objID):
    start_angle_c = mymovehub.last_angle_C
    print('move head')
    try:
        mymovehub.run_motor_for_angle(MOTOR_C, 90, -20)
        sleep(1)
        mymovehub.run_motor_for_angle(MOTOR_C, 270, 20)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_C, 180, -50)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_C, 100, 10)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_C, 100, -10)
        sleep(2)
        mymovehub.return_to_start(MOTOR_C,start_angle_c,mymovehub.last_angle_C)
        sleep(1)
    except:
        print('Cannot connect to Boost.')
def tail (objID):
    print('wiggle tail')
    start_angle_a = mymovehub.last_angle_A
    try:
        mymovehub.run_motor_for_angle(MOTOR_A, 60,  5)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 120, -50)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 120, 50)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 120, -5)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 60, 5)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 60, -50)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 120, 50)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 120, -5)
        sleep(0.5)
        mymovehub.run_motor_for_angle(MOTOR_A, 60,  5)
        sleep(1)
        mymovehub.return_to_start(MOTOR_A, start_angle_a, mymovehub.last_angle_A)
        sleep(1)
    except:
        print('Cannot connect to Boost.')

def legs (objID):
    print('move legs')
    start_angle_c = mymovehub.last_angle_C
    try:
        mymovehub.run_motor_for_angle(MOTOR_B, 180, -20)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_B, 100, -20)
        sleep(1)
        mymovehub.run_motor_for_angle(MOTOR_B, 100, 50)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_B, 180, 20)
        sleep(2)
        mymovehub.return_to_start(MOTOR_B, start_angle_c, mymovehub.last_angle_C)
    except:
        print('Cannot connect to Boost.')


# IMPORTANT: Somewhere:
# mymovehub.stop()
#We then connect them to their trigger object identifiers
objectID2Action = { 50 :head, 52 :tail, 51 :legs}
# create the output object
utopia2output=Utopia2Output(outputPressThreshold=None)
# replace the default action dictionary with ours. N.B. be careful
utopia2output.objectID2Action=objectID2Action
# run the output module
utopia2output.run()



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from mindaffectBCI.utopiaclient import UtopiaClient, Subscribe, StimulusEvent, NewTarget, Selection, DataPacket, UtopiaMessage, SignalQuality
from collections import deque
from mindaffectBCI.decoder.utils import RingBuffer, extract_ringbuffer_segment
from mindaffectBCI.decoder.lower_bound_tracker import lower_bound_tracker
from mindaffectBCI.decoder.linear_trend_tracker import linear_trend_tracker
from time import sleep
import numpy as np

class UtopiaDataInterface:
    """Adaptor class for interfacing between the decoder logic and the data source

    This class provides functionality to wrap a real time data and stimulus stream to make
    it easier to implement standard machine learning pipelines.  In particular it provides streamed
    pre-processing for both EEG and stimulus streams, and ring-buffers for the same with time-stamp based indexing.    
    """


    # TODO [X] : infer valid data time-stamps
    # TODO [X] : smooth and de-jitter the data time-stamps
    # TODO [] : expose a (potentially blocking) message generator interface
    # TODO [X] : ring-buffer for the stimulus-state also, so fast random access
    # TODO [X] : rate limit waiting to reduce computational load
    VERBOSITY = 1

    def __init__(self, datawindow_ms=60000, msgwindow_ms=60000,
                 data_preprocessor=None, stimulus_preprocessor=None, send_signalquality=True, 
                 timeout_ms=100, mintime_ms=50, fs=None, U=None, sample2timestamp='lower_bound_tracker',
                 clientid=None):
        # rate control
        self.timeout_ms = timeout_ms
        self.mintime_ms = mintime_ms # minimum time to spend in update => max processing rate
        # amout of data in the ring-buffer
        self.datawindow_ms = datawindow_ms
        self.msgwindow_ms = msgwindow_ms
        # connect to the mindaffectDecoder
        self.host = None
        self.port = -1
        self.U = UtopiaClient(clientid) if U is None else U
        self.t0 = self.getTimeStamp()
        # init the buffers

        # Messages
        self.msg_ringbuffer = deque()
        self.msg_timestamp = None # ts of most recent processed message

        # DataPackets
        self.data_ringbuffer = None # init later...
        self.data_timestamp = None # ts of last data packet seen
        self.sample2timestamp = sample2timestamp # sample tracker to de-jitter time-stamp information
        self.data_preprocessor = data_preprocessor # function to pre-process the incomming data

        # StimulusEvents
        self.stimulus_ringbuffer = None # init later...
        self.stimulus_timestamp = None # ts of most recent processed data
        self.stimulus_preprocessor = stimulus_preprocessor # function to pre-process the incomming data

        # Info about the data sample rate -- estimated from packet rates..
        self.raw_fs = fs
        self.fs = None
        self.newmsgs = [] # list new unprocssed messages since last update call

        # BODGE: running statistics for sig2noise estimation
        # TODO []: move into it's own Sig2Noise computation class
        self.send_signalquality = send_signalquality
        self.last_sigquality_ts = None
        self.last_log_ts = None
        self.send_sigquality_interval = 1000 # send signal qualities every 1000ms = 1Hz
        # noise2sig estimate halflife_ms, running-offset, de-trended power
        self.noise2sig_halflife_ms = (5000, 500) # 10s for offset, .5s for power
        # TODO [x]: move into a exp-move-ave power est class
        self.raw_power = None
        self.preproc_power = None

    def connect(self, host=None, port=-1, queryifhostnotfound=True):
        '''make a connection to the utopia host'''
        if host:
            self.host = host
        if port > 0:
            self.port = port
        self.U.autoconnect(self.host, self.port, timeout_ms=5000, queryifhostnotfound=queryifhostnotfound)
        if self.U.isConnected:
            # subscribe to messages: data, stim, mode, selection
            self.U.sendMessage(Subscribe(None, "DEMSN"))
        return self.U.isConnected
    
    def isConnected(self):
        return self.U.isConnected if self.U is not None else False

    def getTimeStamp(self):
        '''get the current timeStamp'''
        return self.U.getTimeStamp()

    def sendMessage(self, msg: UtopiaMessage):
        ''' send a UtopiaMessage to the utopia hub'''
        self.U.sendMessage(msg)

    def getNewMessages(self, timeout_ms=0):
        ''' get new messages from the UtopiaHub '''
        return self.U.getNewMessages(timeout_ms)

    def initDataRingBuffer(self):
        '''initialize the data ring buffer, by getting some seed messages and datapackets to get the data sizes etc.'''
        print("geting some initial data to setup the ring buffer")
        # get some initial data to get data shape and sample rate
        databuf = []
        nmsg = 0
        iter = 0
        data_start_ts = None
        data_ts = 0
        while data_start_ts is None or data_ts - data_start_ts < 3000:
            msgs = self.getNewMessages(100)
            for m in msgs:
                m = self.preprocess_message(m)
                if m.msgID == DataPacket.msgID: # data-packets are special
                    if len(m.samples) > 0:
                        databuf.append(m) # append raw data
                        if data_start_ts is None:
                            data_start_ts = m.timestamp
                        data_ts = m.timestamp
                    else:
                        print("Huh? got empty data packet: {}".format(m))
                else:
                    self.msg_ringbuffer.append(m)
                    self.msg_timestamp = m.timestamp
                    nmsg = nmsg+1
        nsamp = [len(m.samples) for m in databuf]
        data_ts = [ m.timestamp for m in databuf]
        if self.raw_fs is None:
            self.raw_fs = np.median( np.array(nsamp[1:])  / np.diff(data_ts) * 1000.0)
        print('Estimated sample rate {} samp in {} s ={}'.format(sum(nsamp),(data_ts[-1]-data_ts[0])/1000.0,self.raw_fs))

        # init the pre-processor (if one)
        if self.data_preprocessor:
            self.data_preprocessor.fit(np.array(databuf[0].samples)[0:1,:], fs=self.raw_fs) # tell it the sample rate

        # apply the data packet pre-processing -- to get the info
        # on the data state after pre-processing
        tmpdatabuf = [self.processDataPacket(m) for m in databuf]
        # strip empty packets
        tmpdatabuf = [d for d in tmpdatabuf if d.shape[0]>0]
        # estimate the sample rate of the pre-processed data
        pp_nsamp = [m.shape[0] for m in tmpdatabuf]
        pp_ts = [ m[-1,-1] for m in tmpdatabuf]
        self.fs = np.median( np.array(pp_nsamp[1:])  / np.diff(pp_ts) * 1000.0)# fs = nSamp/time
        print('Estimated pre-processed sample rate={}'.format(self.fs))

        # create the ring buffer, big enough to store the pre-processed data
        if self.data_ringbuffer:
            print("Warning: re-init data ring buffer")
        # TODO []: why does the datatype of the ring buffer matter so much? Is it because of uss?
        #  Answer[]: it's the time-stamps, float32 rounds time-stamps to 24bits
        self.data_ringbuffer = RingBuffer(maxsize=self.fs*self.datawindow_ms/1000, shape=tmpdatabuf[0].shape[1:], dtype=np.float32)

        # insert the warmup data into the ring buffer
        self.data_timestamp=None # reset last seen data
        nsamp=0
        # re-init the preprocessor for consistency with off-line
        if self.data_preprocessor:
            self.data_preprocessor.fit(np.array(databuf[0].samples)[0:1,:], fs=self.raw_fs)
        # use linear trend tracker to de-jitter the sample timestamps
        if self.sample2timestamp is None or isinstance(self.sample2timestamp,str):
            self.sample2timestamp = timestamp_interpolation(fs=self.fs,
                                                            sample2timestamp=self.sample2timestamp)
        for m in databuf:
            # apply the pre-processing again (this time with fs estimated)
            d = self.processDataPacket(m)
            self.data_ringbuffer.extend(d)
            nsamp = nsamp + d.shape[0]

        return (nsamp, nmsg)

    def initStimulusRingBuffer(self):
        '''initialize the data ring buffer, by getting some seed messages and datapackets to get the data sizes etc.'''
        # TODO []: more efficient memory use, with different dtype for 'real' data and the time-stamps?
        self.stimulus_ringbuffer = RingBuffer(maxsize=self.fs*self.datawindow_ms/1000, shape=(257,), dtype=np.float32)

    def preprocess_message(self, m:UtopiaMessage):
        ''' apply pre-processing to topia message before any more work '''
        #  WARNING BODGE: fit time-stamp in 24bits for float32 ring buffer
        #  Note: this leads to wrap-arroung in (1<<24)/1000/3600 = 4.6 hours
        #        but that shouldn't matter.....
        m.timestamp = m.timestamp % (1<<24)
        return m
    
    def processDataPacket(self, m: DataPacket):
        '''pre-process a datapacket message ready to be inserted into the ringbuffer'''
        #print("DP: {}".format(m))
        # extract the raw data
        d = np.array(m.samples, dtype=np.float32) # process as singles
        # apply the pre-processor, if one was given

        if self.data_preprocessor:
            d_raw = d.copy()
            # warning-- with agressive downsample this may not produce any data!
            d = self.data_preprocessor.transform(d)

            # BODGE: running estimate of the electrode-quality, ONLY after initialization!
            if self.send_signalquality and self.data_ringbuffer is not None:
                self.update_and_send_ElectrodeQualities(d_raw, d, m.timestamp)

                #if self.VERBOSITY > 0 and self.data_ringbuffer is not None:
                #    self.plot_raw_preproc_data(d_raw,d,m.timestamp)

        if d.size > 0 :
            # If have data to add to the ring-buffer, guarding for time-stamp wrap-around
            # TODO [ ]: de-jitter and better timestamp interpolation
            # guard for wrap-around!
            if self.data_timestamp is not None and m.timestamp < self.data_timestamp:
                print("Warning: Time-stamp wrap-around detected!!")

            d = self.add_sample_timestamps(d,m.timestamp,self.fs)

        # update the last time-stamp tracking
        self.data_timestamp= m.timestamp
        return d

    def add_sample_timestamps(self,d:np.ndarray,timestamp:float,fs:float):
        """add per-sample timestamp information to the data matrix

        Args:
            d (np.ndarray): (t,d) the data matrix to attach time stamps to
            timestamp (float): the timestamp of the last sample of d
            fs (float): the nomional sample rate of d

        Returns:
            np.ndarray: (t,d+1) data matrix with attached time-stamp channel
        """
        if self.sample2timestamp is not None and not isinstance(self.sample2timestamp,str):
            sample_ts = self.sample2timestamp.transform(timestamp, len(d))
        else: # all the same ts
            sample_ts = np.ones((len(d),),dtype=int)*timestamp
        # combine data with timestamps, ensuring type is preserved
        d = np.append(np.array(d), sample_ts[:, np.newaxis], -1).astype(d.dtype)
        return d

    def plot_raw_preproc_data(self, d_raw, d_preproc, ts):
        '''debugging function to check the diff between the raw and pre-processed data'''
        if not hasattr(self,'rawringbuffer'):
            self.preprocringbuffer=RingBuffer(maxsize=self.fs*3,shape=(d_preproc.shape[-1]+1,))
            self.rawringbuffer=RingBuffer(maxsize=self.raw_fs*3,shape=(d_raw.shape[-1]+1,))
        d_preproc = self.add_sample_timestamps(d_preproc,ts,self.fs)
        self.preprocringbuffer.extend(d_preproc)
        d_raw = self.add_sample_timestamps(d_raw,ts,self.raw_fs)
        self.rawringbuffer.extend(d_raw)
        if self.last_sigquality_ts is None or ts > self.last_sigquality_ts + self.send_sigquality_interval:
            import matplotlib.pyplot as plt
            plt.figure(10);plt.clf();
            idx = np.flatnonzero(self.rawringbuffer[:,-1])[0]
            plt.subplot(211); plt.cla(); plt.plot(self.rawringbuffer[idx:,-1],self.rawringbuffer[idx:,:-1])
            idx = np.flatnonzero(self.preprocringbuffer[:,-1])[0]
            plt.subplot(212); plt.cla(); plt.plot(self.preprocringbuffer[idx:,-1],self.preprocringbuffer[idx:,:-1])
            plt.show(block=False)


    def processStimulusEvent(self, m: StimulusEvent):
        '''pre-process a StimulusEvent message ready to be inserted into the stimulus ringbuffer'''
        # get the vector to hold the stimulus info
        d = np.zeros((257,),dtype=np.float32)

        if self.stimulus_ringbuffer is not None and self.stimulus_timestamp is not None:
            # hold value of used objIDs from previous time stamp
            d[:] = self.stimulus_ringbuffer[-1,:]

        # insert the  updated state
        d[m.objIDs] = m.objState
        d[-1] = m.timestamp
        # apply the pre-processor, if one was given
        if self.stimulus_preprocessor:
            d = self.stimulus_preprocessor.transform(d)

        # update the last time-stamp tracking
        self.stimulus_timestamp= m.timestamp
        return d

    def update_and_send_ElectrodeQualities(self, d_raw: np.ndarray, d_preproc: np.ndarray, ts: int):
        ''' compute running estimate of electrode qality and stream it '''
        raw_power, preproc_power = self.update_electrode_powers(d_raw, d_preproc)

        # convert to average amplitude
        raw_amp = np.sqrt(raw_power)
        preproc_amp = np.sqrt(preproc_power)

        # noise2signal estimated as removed raw amplitude (assumed=noise) to preprocessed amplitude (assumed=signal)
        noise2sig = np.maximum(float(1e-6), np.abs(raw_amp - preproc_amp)) /  np.maximum(float(1e-8),preproc_amp)

        # hack - detect disconnected channels
        noise2sig[ raw_power < 1e-6 ] = 100

        # hack - detect filter artifacts = preproc power is too big..
        noise2sig[ preproc_amp > raw_amp*10 ] = 100

        # hack - cap to 100
        noise2sig = np.minimum(noise2sig,100)

        # rate limit sending of signal-quality messages
        if self.last_sigquality_ts is None or ts > self.last_sigquality_ts + self.send_sigquality_interval:
            print("SigQ:\nraw_power=({}/{})\npp_power=({}/{})\nnoise2sig={}".format(
                   raw_amp,d_raw.shape[0],
                   preproc_amp,d_preproc.shape[0],
                   noise2sig))
            print("Q",end='')
            # N.B. use *our* time-stamp for outgoing messages!
            self.sendMessage(SignalQuality(None, noise2sig))
            self.last_sigquality_ts = ts

            if self.VERBOSITY>2:
                # plot the sample time-stamp jitter...
                import matplotlib.pyplot as plt
                plt.figure(10)
                ts = self.data_ringbuffer[:,-1]
                idx = np.flatnonzero(ts)
                if len(idx)>0:
                    ts = ts[idx[0]:]
                    plt.subplot(211); plt.cla(); plt.plot(np.diff(ts)); plt.title('diff time-sample')
                    plt.subplot(212); plt.cla(); plt.plot((ts-ts[0])-np.arange(len(ts))*1000.0/self.fs); plt.title('regression against sample-number')
                    plt.show(block=False)

    def update_electrode_powers(self, d_raw: np.ndarray, d_preproc:np.ndarray):
        ''' track exp-weighted-moving average centered power for 2 input streams '''
        if self.raw_power is None:
            mu_hl, pow_hl = self.noise2sig_halflife_ms
            self.raw_power = power_tracker(mu_hl, pow_hl, self.raw_fs)
            self.preproc_power = power_tracker(mu_hl, pow_hl, self.fs)
        self.raw_power.transform(d_raw)
        self.preproc_power.transform(d_preproc)
        return (self.raw_power.power(), self.preproc_power.power())


    def update(self, timeout_ms=None, mintime_ms=None):
        '''Update the tracking state w.r.t. the data source

        By adding data to the data_ringbuffer, stimulus info to the stimulus_ringbuffer, 
        and other messages to the messages ring buffer.

        Args
         timeout_ms : int
             max block waiting for messages before returning
         mintime_ms : int
             min time to accumulate messages before returning
        Returns
          newmsgs : [newMsgs :UtopiaMessage]
             list of the *new* utopia messages from the server
          nsamp: int
             number of new data samples in this call
             Note: use data_ringbuffer[-nsamp:,...] to get the new data
          nstimulus : int
             number of new stimulus events in this call
             Note: use stimulus_ringbuffer[-nstimulus:,...] to get the new data
        '''
        if timeout_ms is None:
            timeout_ms = self.timeout_ms
        if mintime_ms is None:
            mintime_ms = self.mintime_ms
        if not self.isConnected():
            self.connect()
        if not self.isConnected():
            return [],0,0

        t0 = self.getTimeStamp()
        nsamp = 0
        nmsg = 0
        nstimulus = 0

        if self.data_ringbuffer is None: # do special init stuff if not done
            nsamp, nmsg = self.initDataRingBuffer()
        if self.stimulus_ringbuffer is None: # do special init stuff if not done
            self.initStimulusRingBuffer()
        if self.last_log_ts is None:
            self.last_log_ts = self.getTimeStamp()
        if t0 is None:
            t0 = self.getTimeStamp()

        # record the list of new messages from this call
        newmsgs = self.newmsgs # start with any left-overs from old calls 
        self.newmsgs=[] # clear the  left-over messages stack
        
        ttg = timeout_ms - (self.getTimeStamp() - t0) # time-to-go in the update loop
        while ttg > 0:

            # rate limit
            if ttg >= mintime_ms:
                sleep(mintime_ms/1000.0)
                ttg = timeout_ms - (self.getTimeStamp() - t0) # udate time-to-go
                
            # get the new messages
            msgs = self.getNewMessages(ttg)

            # process the messages - basically to split datapackets from the rest
            print(".",end='')
            #print("{} in {}".format(len(msgs),self.getTimeStamp()-t0),end='',flush=True)
            for m in msgs:
                m = self.preprocess_message(m)
                
                print("{:c}".format(m.msgID), end='', flush=True)
                
                if m.msgID == DataPacket.msgID: # data-packets are special
                    d = self.processDataPacket(m) # (samp x ...)
                    self.data_ringbuffer.extend(d)
                    nsamp = nsamp + d.shape[0]
                    
                elif m.msgID == StimulusEvent.msgID: # as are stmiuluse events
                    d = self.processStimulusEvent(m) # (nY x ...)
                    self.stimulus_ringbuffer.append(d)
                    nstimulus = nstimulus + 1
                    
                else:
                    # NewTarget/Selection are also special in that they clear stimulus state...
                    if m.msgID == NewTarget.msgID or m.msgID == Selection.msgID :
                        # Make a dummy stim-event to reset all objIDs to off
                        d = self.processStimulusEvent(StimulusEvent(m.timestamp,
                                                                    np.arange(255,dtype=np.int32),
                                                                    np.zeros(255,dtype=np.int8)))
                        self.stimulus_ringbuffer.append(d)
                        self.stimulus_timestamp= m.timestamp
                    
                    if len(self.msg_ringbuffer)>0 and m.timestamp > self.msg_ringbuffer[0].timestamp + self.msgwindow_ms: # slide msg buffer
                        self.msg_ringbuffer.popleft()
                    self.msg_ringbuffer.append(m)
                    newmsgs.append(m)
                    nmsg = nmsg+1
                    self.msg_timestamp = m.timestamp
                
            # update time-to-go
            ttg = timeout_ms - (self.getTimeStamp() - t0)

        # new line
        if self.getTimeStamp() > self.last_log_ts + 2000:
            print("",flush=True)
            self.last_log_ts = self.getTimeStamp()
        
        # return new messages, and count new samples/stimulus 
        return (newmsgs, nsamp, nstimulus)

    def push_back_newmsgs(self,oldmsgs):
        '''put unprocessed messages back onto the newmessages queue'''
        # TODO []: ensure  this preserves message time-stamp order?
        self.newmsgs.extend(oldmsgs)

    def extract_data_segment(self, bgn_ts, end_ts=None):
        """extract a segment of data based on a start and end time-stamp

        Args:
            bgn_ts (float): segment start time-stamp
            end_ts (float, optional): segment end time-stamp. Defaults to None.

        Returns:
            (np.ndarray): the data between these time-stamps, or None if timestamps invalid
        """        
        return extract_ringbuffer_segment(self.data_ringbuffer,bgn_ts,end_ts)
    
    def extract_stimulus_segment(self, bgn_ts, end_ts=None):
        """extract a segment of the stimulus stream based on a start and end time-stamp

        Args:
            bgn_ts (float): segment start time-stamp
            end_ts (float, optional): segment end time-stamp. Defaults to None.

        Returns:
            (np.ndarray): the stimulus events between these time-stamps, or None if timestamps invalid
        """        
        return extract_ringbuffer_segment(self.stimulus_ringbuffer,bgn_ts,end_ts)
    
    def extract_msgs_segment(self, bgn_ts, end_ts=None):
        ''' extract the messages between start/end time stamps'''
        msgs = [] # store the trial stimEvents
        for m in reversed(self.msg_ringbuffer):
            if m.timestamp <= bgn_ts:
                # stop as soon as earlier than bgn_ts
                break
            if end_ts is None or m.timestamp < end_ts:
                msgs.append(m)
        # reverse back to input order
        msgs.reverse()
        return msgs

    def run(self, timeout_ms=30000):
        '''test run the interface forever, just getting and storing data'''
        t0 = self.getTimeStamp()
        # test getting 5s data
        tstart = self.data_timestamp
        trlen_ms = 5000
        while self.getTimeStamp() < t0+timeout_ms:
            self.update()
            # test getting a data segment
            if tstart is None :
                tstart = self.data_timestamp
            if tstart and self.data_timestamp > tstart + trlen_ms:
                X = self.extract_data_segment(tstart, tstart+trlen_ms)
                print("Got data: {}->{}\n{}".format(tstart, tstart+trlen_ms, X[:, -1]))
                Y = self.extract_stimulus_segment(tstart, tstart+trlen_ms)
                print("Got stimulus: {}->{}\n{}".format(tstart, tstart+trlen_ms, Y[:, -1]))
                tstart = self.data_timestamp + 5000
            print('.', flush=True)


try:
    from sklearn.base import TransformerMixin
except:
    # fake the class if sklearn is not available, e.g. Android/iOS
    class TransformerMixin:
        def __init__():
            pass
        def fit(self,X):
            pass
        def transform(self,X):
            pass






#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
from mindaffectBCI.decoder.utils import sosfilt, butter_sosfilt, sosfilt_zi_warmup
class butterfilt_and_downsample(TransformerMixin):
    """Incremental streaming transformer to provide filtering and downsampling data transformations

    Args:
        TransformerMixin ([type]): sklearn compatible transformer
    """    
    def __init__(self, stopband=((0,5),(5,-1)), order:int=6, fs:float =250, fs_out:float =60, ftype='butter'):
        self.stopband = stopband
        self.fs = fs
        self.fs_out = fs_out if fs_out is not None and fs_out < fs else fs
        self.order = order
        self.axis = -2
        if not self.axis == -2:
            raise ValueError("axis != -2 is not yet supported!")
        self.nsamp = 0
        self.ftype = ftype

    def fit(self, X, fs:float =None, zi=None):
        if fs is not None: # parameter overrides stored fs
            self.fs = fs

        # preprocess -> spectral filter
        if isinstance(self.stopband, str):
            import pickle
            import os
            # load coefficients from file -- when scipy isn't available
            if os.path.isfile(self.stopband):
                fn = self.stopband 
            else: # try relative to our py file
                fn = os.path.join(os.path.dirname(os.path.abspath(__file__)),self.stopband)
            with open(fn,'rb') as f:
                self.sos_ = pickle.load(f)
                self.zi_ = pickle.load(f)
                f.close()
            # tweak the shape/scale of zi to the actual data shape
            self.zi_ = sosfilt_zi_warmup(self.zi_, X, self.axis)
            print("X={} zi={}".format(X.shape,self.zi_.shape))

        else:
            # estimate them from the given information
            X, self.sos_, self.zi_ = butter_sosfilt(X, self.stopband, self.fs, order=self.order, axis=self.axis, zi=zi, ftype=self.ftype)
            
        # preprocess -> downsample
        self.nsamp = 0
        self.resamprate_ = int(round(self.fs*2.0/self.fs_out))/2.0 if self.fs_out is not None else 1
        self.out_fs_  = self.fs/self.resamprate_
        print("resample: {}->{}hz rsrate={}".format(self.fs, self.out_fs_, self.resamprate_))

        return self

    def transform(self, X, Y=None):
        # propogate the filter coefficients between calls
        if not hasattr(self,'sos_'):
            self.fit(X[0:1,:])

        if self.sos_ is not None:
            X, self.zi_ = sosfilt(self.sos_, X, axis=self.axis, zi=self.zi_)

        nsamp = self.nsamp
        self.nsamp = self.nsamp + X.shape[self.axis] # track *raw* sample counter

        # preprocess -> downsample @60hz
        if self.resamprate_ > 1:
            # number samples through this cycle due to remainder of last block
            resamp_start = nsamp%self.resamprate_
            # convert to number samples needed to complete this cycle
            # this is then the sample to take for the next cycle
            if resamp_start > 0:
                resamp_start = self.resamprate_ - resamp_start
            
            # allow non-integer resample rates
            idx =  np.arange(resamp_start,X.shape[self.axis],self.resamprate_)

            if self.resamprate_%1 > 0 and idx.size>0 : # non-integer re-sample, interpolate
                idx_l = np.floor(idx).astype(int) # sample above
                idx_u = np.ceil(idx).astype(int) # sample below
                # BODGE: guard for packet ending at sample boundary.
                idx_u[-1] = idx_u[-1] if idx_u[-1]<X.shape[self.axis] else X.shape[self.axis]-1
                w_u   = idx - idx_l # linear weight of the upper sample
                X = X[...,idx_u,:] * w_u[:,np.newaxis] + X[...,idx_l,:] * (1-w_u[:,np.newaxis]) # linear interpolation
                if Y is not None:
                    Y = Y[...,idx_u,:] * w_u[:,np.newaxis] + Y[...,idx_l,:] * (1-w_u[:,np.newaxis])

            else:
                idx = idx.astype(int)
                X = X[..., idx, :] # decimate X (trl, samp, d)
                if Y is not None:
                    Y = Y[..., idx, :] # decimate Y (trl, samp, y)
        
        return X if Y is None else (X, Y)

    @staticmethod
    def testcase():
        ''' test the filt+downsample transformation filter by incremental calling '''
        #X=np.cumsum(np.random.randn(100,1),axis=0)
        X=np.sin(np.arange(100)[:,np.newaxis]*2*np.pi/30)
        xs = np.arange(X.shape[0])[:,np.newaxis]
        # high-pass and decimate
        bands = ((0,20,'bandpass'))
        fs = 200
        fs_out = 130
        fds = butterfilt_and_downsample(stopband=bands,fs=fs,fs_out=fs_out)

        
        print("single step")
        fds.fit(X[0:1,:])
        m0,xs0 = fds.transform(X,xs) # (samp,ny,ne)
        print("M0 -> {}".format(m0[:20]))

        step=6
        print("Step size = {}".format(step))
        fds.fit(X[0:1,:])
        m1=np.zeros(m0.shape,m0.dtype)
        xs1 = np.zeros(xs0.shape,xs0.dtype)
        t=0
        for i in range(0,len(X),step):
            idx=np.arange(i,min(i+step,len(X)))
            mm, idx1=fds.transform(X[idx,:],idx[:,np.newaxis])
            m1[t:t+mm.shape[0],:]=mm
            xs1[t:t+mm.shape[0]]=idx1
            t = t +mm.shape[0]
        print("M1 -> {}".format(m1[:20]))
        print("diff: {}".format(np.max(np.abs(m0-m1))))

        import matplotlib.pyplot as plt 
        plt.plot(xs,X,'*-',label='X')
        plt.plot(xs0,m0,'*-',label='{} {}->{}Hz single'.format(bands,fs,fs_out))
        plt.plot(xs1,m1,'*-',label='{} {}->{}Hz incremental'.format(bands,fs,fs_out))
        plt.legend()
        plt.show()





#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
from mindaffectBCI.decoder.stim2event import stim2event
class stim2eventfilt(TransformerMixin):
    ''' Incremental streaming transformer to transform a sequence of stimulus states to a brain event sequence
    
    For example by transforming a sequence of stimulus intensities, to rising and falling edge events.
    '''
    def __init__(self, evtlabs=None, histlen=20):
        self.evtlabs = evtlabs
        self.histlen = histlen
        self.prevX = None

    def fit(self, X):
        return self

    def transform(self, X):
        '''transform Stimulus-encoded to brain-encoded'''
        if X is None:
            return None
        
        # keep old fitler state for the later transformation call
        prevX = self.prevX

        # grab the new filter state (if wanted)
        if self.histlen>0:
            #print('prevX={}'.format(prevX))
            #print("X={}".format(X))
            if X.shape[0] >= self.histlen or prevX is None:
                self.prevX = X
            else:
                self.prevX = np.append(prevX, X, 0)
            # only keep the last bit -- copy in case gets changed in-place
            self.prevX = self.prevX[-self.histlen:,:].copy()
            #print('new_prevX={}'.format(self.prevX))

        # convert from stimulus coding to brain response coding, with old state
        X = stim2event(X, self.evtlabs, axis=-2, oM=prevX)
        return X

    def testcase():
        ''' test the stimulus transformation filter by incremental calling '''
        M=np.array([0,0,0,1,0,0,1,1,0,1])[:,np.newaxis] # samp,nY
        s2ef = stim2eventfilt(evtlabs=('re','fe'),histlen=3)

        print("single step")
        m0=s2ef.transform(M) # (samp,ny,ne)
        print("{} -> {}".format(M,m0))

        print("Step size = 1")
        m1=np.zeros(m0.shape,m0.dtype)
        for i in range(len(M)):
            idx=slice(i,i+1)
            mm=s2ef.transform(M[idx,:])
            m1[idx,...]=mm
            print("{} {} -> {}".format(i,M[idx,...],mm))

        print("Step size=4")
        m4=np.zeros(m0.shape,m0.dtype)
        for i in range(0,len(M),4):
            idx=slice(i,i+4)
            mm=s2ef.transform(M[idx,:])
            m4[idx,...]=mm
            print("{} {} -> {}".format(i,M[idx,...],mm))

        print("m0={}\nm1={}\n,m4={}\n".format(m0,m1,m4))
            




#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
class power_tracker(TransformerMixin):
    """Incremental streaming transformer from raw n-channel data, to exponientially smoothed channel powers

    Args:
        TransformerMixin ([type]): sklearn compatiable transformer
    """

    def __init__(self,halflife_mu_ms, halflife_power_ms, fs):
        # convert to per-sample decay factor
        self.alpha_mu = self.hl2alpha(fs * halflife_mu_ms / 1000.0 ) 
        self.alpha_power= self.hl2alpha(fs * halflife_power_ms / 1000.0 )
        self.sX_N = None
        self.sX = None
        self.sXX_N = None
        self.sXX = None

    def hl2alpha(self,hl):
        return np.exp(np.log(.5)/hl)

    def fit(self,X):
        self.sX_N = X.shape[0]
        self.sX = np.sum(X,axis=0)
        self.sXX_N = X.shape[0]
        self.sXX = np.sum((X-(self.sX/self.sX_N))**2,axis=0)
        return self.power()

    def transform(self, X: np.ndarray):
        ''' compute the exponientially weighted centered power of X '''
        if self.sX is None: # not fitted yet!
            return self.fit(X)
        # compute updated mean
        alpha_mu   = self.alpha_mu ** X.shape[0]
        self.sX_N  = self.sX_N*alpha_mu + X.shape[0]
        self.sX    = self.sX*alpha_mu + np.sum(X, axis=0)
        # center and compute updated power
        alpha_pow  = self.alpha_power ** X.shape[0]
        self.sXX_N = self.sXX_N*alpha_pow + X.shape[0]
        self.sXX   = self.sXX*alpha_pow + np.sum((X-(self.sX/self.sX_N))**2, axis=0)       
        return self.power()
    
    def mean(self):
        return self.sX / self.sX_N
    def power(self):
        return self.sXX / self.sXX_N
    
    def testcase(self):
        X = np.random.randn(10000,2)
        #X = np.cumsum(X,axis=0)
        pt = power_tracker(100,100,100)
        print("All at once: power={}".format(pt.transform(X)))  # all at once
        pt = power_tracker(100,1000,1000)
        print("alpha_mu={} alpha_pow={}".format(pt.alpha_mu,pt.alpha_power) )
        step = 30
        idxs = list(range(step,X.shape[0],step))
        powers = np.zeros((len(idxs),X.shape[-1]))
        mus = np.zeros((len(idxs),X.shape[-1]))
        for i,j in enumerate(idxs):
            powers[i,:] = np.sqrt(pt.transform(X[j-step:j,:]))
            mus[i,:]=pt.mean()
        for d in range(X.shape[-1]):
            plt.subplot(X.shape[-1],1,d+1)
            plt.plot(X[:,d])
            plt.plot(idxs,mus[:,d])
            plt.plot(idxs,powers[:,d])




#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
class timestamp_interpolation(TransformerMixin):
    """Incremental streaming tranformer to transform from per-packet time-stamps to per-sample timestamps 
    with time-stamp smoothing, de-jittering, and dropped-sample detection.
    """

    def __init__(self,fs=None,sample2timestamp=None, max_delta=200):
        """tranform from per-packet (i.e. multiple-samples) to per-sample timestamps

        Args:
            fs (float): default sample rate, used when no other timing info is available
            sample2timestamp (transformer, optional): class to de-jitter timestamps based on sample-count. Defaults to None.
        """        
        self.fs=fs
        a0 = 1000/self.fs if self.fs is not None else 1
         # BODGE: special cases for particular mapping functions so can include the prior slope
        if sample2timestamp=='lower_bound_tracker':
            self.sample2timestamp = lower_bound_tracker(a0=a0)
        elif sample2timestamp=='linear_trend_tracker':
            self.sample2timestamp = linear_trend_tracker(a0=a0)
        else:
            self.sample2timestamp = sample2timestamp
        self.max_delta = max_delta

    def fit(self,ts,nsamp=1):
        self.last_sample_timestamp_ = ts
        self.n_ = 0

    def transform(self,timestamp:float,nsamp:int=1):
        """add per-sample timestamp information to the data matrix

        Args:
            timestamp (float): the timestamp of the last sample of d
            nsamp(int): number of samples to interpolate

        Returns:
            np.ndarray: (nsamp) the interpolated time-stamps
        """
        if not hasattr(self,'last_sample_timestamp_'):
            self.fit(timestamp,nsamp)

        # update tracking number samples processed
        self.n_ = self.n_ + nsamp

        if self.last_sample_timestamp_ < timestamp or self.sample2timestamp is not None:
            # update the tracker for the sample-number to sample timestamp mapping
            if self.sample2timestamp is not None:
                #print("n={} ts={}".format(n,timestamp))
                newtimestamp = self.sample2timestamp.transform(self.n_, timestamp)
                #print("ts={} newts={} diff={}".format(timestamp,newtimestamp,timestamp-newtimestamp))
                # use the corrected de-jittered time-stamp -- if it's not tooo different
                if abs(timestamp-newtimestamp) < self.max_delta:
                    timestamp = int(newtimestamp)

            # simple linear interpolation for the sample time-stamps
            samples_ts = np.linspace(self.last_sample_timestamp_, timestamp, nsamp+1, endpoint=True, dtype=int)
            samples_ts = samples_ts[1:]
        else:
            if self.fs :
                # interpolate with the estimated sample rate                    
                samples_ts = np.arange(-nsamp+1,1,dtype=int)*(1000/self.fs) + timestamp
            else:
                # give all same timestamp
                samples_ts = np.ones(nsamp,dtype=int)*timestamp

        # update the tracking info
        self.last_sample_timestamp_ = timestamp

        return samples_ts

    def testcase(self, npkt=1000, fs=100):
        # generate random packet sizes
        nsamp = np.random.random_integers(0,10,size=(npkt,))
        # generate true sample timestamps
        ts_true = np.arange(np.sum(nsamp))*1000/fs
        # packet end indices
        idx = np.cumsum(nsamp)-1
        # packet end time-stamps
        pkt_ts = ts_true[idx]
        # add some time-stamp jitter, always positive..
        pkt_ts = pkt_ts + np.random.uniform(0,.5*1000/fs,size=pkt_ts.shape)
        # apply the time-stamp interplotation
        sts=[]
        tsfn = timestamp_interpolation(fs=fs,sample2timestamp = 'lower_bound_tracker')
        for i,(n,t) in enumerate(zip(nsamp,pkt_ts)):
            samp_ts = tsfn.transform(t,n)
            sts.extend(samp_ts)
        # plot the result.
        import matplotlib.pyplot as plt
        plt.plot(ts_true - sts)
        plt.show()

def testRaw():
    # test with raw
    ui = UtopiaDataInterface()
    ui.connect()
    sigViewer(ui,30000) # 30s sigviewer

def testPP():
    from sigViewer import sigViewer
    # test with a filter + downsampler
    ppfn= butterfilt_and_downsample(order=4, stopband=((0,1),(25,-1)), fs_out=100)
    #ppfn= butterfilt_and_downsample(order=4, stopband='butter_stopband((0, 5), (25, -1))_fs200.pk', fs_out=80)     
    ui = UtopiaDataInterface(data_preprocessor=ppfn, stimulus_preprocessor=None)
    ui.connect()
    sigViewer(ui)

def testFileProxy(filename,fs_out=999):
    from mindaffectBCI.decoder.FileProxyHub import FileProxyHub
    U = FileProxyHub(filename)
    from sigViewer import sigViewer
    # test with a filter + downsampler
    #ppfn= butterfilt_and_downsample(order=4, stopband=((0,3),(25,-1)), fs_out=fs_out)
    ppfn= butterfilt_and_downsample(order=4, stopband=(1,15,'bandpass'), fs_out=fs_out)
    #ppfn = None
    ui = UtopiaDataInterface(data_preprocessor=ppfn, stimulus_preprocessor=None, mintime_ms=0, U=U)
    ui.connect()
    sigViewer(ui)

def testFileProxy2(filename):
    from mindaffectBCI.decoder.FileProxyHub import FileProxyHub
    U = FileProxyHub(filename)
    fs = 200
    fs_out = 200
    # test with a filter + downsampler
    ppfn= butterfilt_and_downsample(order=4, stopband=((45,65),(0,3),(25,-1)), fs=fs, fs_out=fs_out)
    ui = UtopiaDataInterface(data_preprocessor=ppfn, stimulus_preprocessor=None, mintime_ms=0, U=U, fs=fs)
    ui.connect()
    # run in bits..
    data=[]
    stim=[]
    emptycount = 0 
    while True:
        newmsg, nsamp, nstim = ui.update()
        if len(newmsg) == 0 and nsamp == 0 and nstim == 0: 
            emptycount = emptycount + 1
            if emptycount > 10:
                break
        else:
            emptycount=0
        if nsamp > 0:
            data.append(ui.data_ringbuffer[-nsamp:,:].copy())
        if nstim > 0:
            stim.append(ui.stimulus_ringbuffer[-nstim:,:].copy())
    # convert to single data block
    data = np.vstack(data)
    stim = np.vstack(stim)
    # dump as pickle
    import pickle
    if ppfn is None:
        pickle.dump(dict(data=data,stim=stim),open('raw_udi.pk','wb'))
    else:
        pickle.dump(dict(data=data,stim=stim),open('pp_udi.pk','wb'))

def testERP():
    ui = UtopiaDataInterface()
    ui.connect()
    erpViewer(ui,evtlabs=None) # 30s sigviewer

def testElectrodeQualities(X,fs=200,pktsize=20):
    # recurse if more dims than we want...
    if X.ndim>2:
        sigq=[]
        for i in range(X.shape[0]):
            sigqi = testElectrodeQualities(X[i,...],fs,pktsize)
            sigq.append(sigqi)
        sigq=np.concatenate(sigq,0)
        return sigq
    
    ppfn= butterfilt_and_downsample(order=6, stopband='butter_stopband((0, 5), (25, -1))_fs200.pk', fs_out=100)
    ppfn.fit(X[:10,:],fs=200)
    noise2sig = np.zeros((int(X.shape[0]/pktsize),X.shape[-1]),dtype=np.float32)
    for pkti in range(noise2sig.shape[0]):
        t = pkti*pktsize
        Xi = X[t:t+pktsize,:]
        Xip = ppfn.transform(Xi)
        raw_power, preproc_power = UtopiaDataInterface.update_electrode_powers(Xi,Xip)
        noise2sig[pkti,:] = np.maximum(float(1e-6), (raw_power - preproc_power)) /  np.maximum(float(1e-8),preproc_power)
    return noise2sig

    
if __name__ == "__main__":
    #timestamp_interpolation().testcase()
    #butterfilt_and_downsample.testcase()
    #testRaw()
    #testPP()
    #testERP()
    filename="C:/Users/Developer/Downloads/mindaffectBCI__201002_1713.txt"
    filename="C:/Users/Developer/Downloads/trig_check/mindaffectBCI_cyton_brainflow_201002_2008.txt"
    testFileProxy(filename)
    #testFileProxy2(filename)
    # "C:\\Users\\Developer\\Downloads\\mark\\mindaffectBCI_brainflow_200911_1229_90cal.txt")
    #"..\..\Downloads\khash\mindaffectBCI_noisetag_bci_200907_1433.txt"



import subprocess
import os

def run(label:str='', host:str='-', usb_port:str="com4", config=8, use_aux:int=0, serial_event:int=0, packet_size:int=5, **kwargs):
    pydir = os.path.dirname(os.path.abspath(__file__)) # mindaffectBCI/examples/acquisation
    bindir = os.path.join(pydir,'..','..','..','bin') # ../../../bin/

    # command to run the java hub
    classpath = ";".join((os.path.join(bindir,"UtopiaServer.jar"),"jssc-2.9.2.jar","openBCI2utopia.jar"))
    cmd = ("java","-cp", classpath, "openBCI2utopia")
    # args to pass to the java hub
    args = (usb_port, host, str(config), str(use_aux), str(serial_event), str(packet_size))

    # run the command, waiting until it has finished
    print("Running command: {}".format(cmd+args))
    acquisation = subprocess.run(cmd + args, cwd=pydir, shell=False,
                               stdin=subprocess.DEVNULL)
    print("Result: {}".format(acquisation.stdout))
    return acquisation

if __name__=="__main__":
    run()


#!/usr/bin/env python3

#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Set up imports and paths
import time
from mindaffectBCI.noisetag import Noisetag
from gpiozero import LED 


nt=None
leds=[]
objIDs=[]
framerate_hz = 60

#---------------------------------------------------------------------
def draw():
    """draw the LED state"""
    #print("Background state"+str(backgroundstate))
    # get the updated stimulus state info
    global nt, leds, framerate
    nt.updateStimulusState()
    stimulus_state,target_state,objIDs,sendEvents=nt.getStimulusState()

    # BODGE: sleep to limit the stimulus update rate
    time.sleep(1/framerate_hz)
    # update the state of each LED to match the stimulusstate
    for i,led in enumerate(leds): 
        # get the background state of this cell
        bs = stimulus_state[i] if stimulus_state else None
        if not bs is None and bs>0 :
            led.on()
        else :
            led.off()

    if target_state is not None and target_state>=0:
        print("*" if target_state==1 else '.', end='', flush=True)
    # send info on updated display state
    nt.sendStimulusState()

def selectionHandler(objID):
    print("Selection: objID=%d"%(objID))

#------------------------------------------------------------------------
# Initialization : display
def init(numleds=2, led2gpiopin=(2,3,4), nCal=10, nPred=10):
    global nt, objIDs, leds
    
    if led2gpiopin is None:
        led2gpiopin = list(range(numleds))
    leds=[]
    objIDs=[]
    for i in range(numleds):
        leds.append(LED(led2gpiopin[i]))
        objIDs.append(i+1)

    nt=Noisetag()
    nt.connect()
    nt.setActiveObjIDs(objIDs)
    nt.startExpt(nCal=nCal,nPred=nPred,
                cueduration=4,duration=10,feedbackduration=4)
    # register function to call if selection is made
    nt.addSelectionHandler(selectionHandler)

if __name__=="__main__":
    framerate_hz = 15
    init(numleds=1, led2gpiopin=(2,3,4), nCal=10, nPred=100)
    while True :
        draw()



import numpy as np
from mindaffectBCI.decoder.datasets import get_dataset
from mindaffectBCI.decoder.model_fitting import BaseSequence2Sequence, MultiCCA, FwdLinearRegression, BwdLinearRegression, LinearSklearn
from sklearn.linear_model import Ridge, LogisticRegression
from sklearn.svm import LinearSVR, LinearSVC
from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, plot_erp, plot_summary_statistics, plot_factoredmodel
from mindaffectBCI.decoder.scoreStimulus import factored2full, plot_Fe
from mindaffectBCI.decoder.decodingCurveSupervised import decodingCurveSupervised, print_decoding_curve, plot_decoding_curve
from mindaffectBCI.decoder.scoreOutput import plot_Fy
from mindaffectBCI.decoder.preprocess import preprocess, plot_grand_average_spectrum
import matplotlib.pyplot as plt
import gc
import re

def analyse_dataset(X:np.ndarray, Y:np.ndarray, coords, model:str='cca', cv=True, tau_ms:float=300, fs:float=None,  rank:int=1, evtlabs=None, offset_ms=0, center=True, **kwargs):
    """ cross-validated training on a single datasets and decoing curve estimation

    Args:
        X (np.ndarray): the X (EEG) sequence
        Y (np.ndarray): the Y (stimulus) sequence
        coords ([type]): array of dicts of meta-info describing the structure of X and Y
        fs (float): the data sample rate (if coords is not given)
        model (str, optional): The type of model to fit, as in `model_fitting.py`. Defaults to 'cca'.
        cv (bool, optional): flag if we should train with cross-validation using the cv_fit method. Defaults to True.
        tau_ms (float, optional): length of the stimulus-response in milliseconds. Defaults to 300.
        rank (int, optional): rank of the decomposition in factored models such as cca. Defaults to 1.
        evtlabs ([type], optional): The types of events to used to model the brain response, as used in `stim2event.py`. Defaults to None.
        offset_ms ((2,):float, optional): Offset for analysis window from start/end of the event response. Defaults to 0.

    Raises:
        NotImplementedError: if you as for a model which isn't implemented

    Returns:
        score (float): the cv score for this dataset
        dc (tuple): the information about the decoding curve as returned by `decodingCurveSupervised.py`
        Fy (np.ndarray): the raw cv'd output-scores for this dataset as returned by `decodingCurveSupervised.py` 
        clsfr (BaseSequence2Sequence): the trained classifier
    """
    # extract dataset info
    if coords is not None:
        fs = coords[1]['fs'] 
        print("X({})={}, Y={} @{}hz".format([c['name'] for c in coords], X.shape, Y.shape, fs))
    else:
        print("X={}, Y={} @{}hz".format(X.shape, Y.shape, fs))
    tau = int(tau_ms*fs/1000)
    offset=int(offset_ms*fs/1000)

    Cscale = np.sqrt(np.mean(X.ravel()**2))
    print('Cscale={}'.format(Cscale))
    C = .1/Cscale

    # create the model if not provided
    if isinstance(model,BaseSequence2Sequence):
        clsfr = model
    elif model=='cca' or model is None:
        clsfr = MultiCCA(tau=tau, offset=offset, rank=rank, evtlabs=evtlabs, center=center, **kwargs)
    elif model=='bwd':
        clsfr = BwdLinearRegression(tau=tau, offset=offset, evtlabs=evtlabs, center=center, **kwargs)
    elif model=='fwd':
        clsfr = FwdLinearRegression(tau=tau, offset=offset, evtlabs=evtlabs, center=center, **kwargs)
    elif model == 'ridge': # should be equivalent to BwdLinearRegression
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, clsfr=Ridge(alpha=0,fit_intercept=center), **kwargs)
    elif model == 'lr':
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, clsfr=LogisticRegression(C=C,fit_intercept=center), labelizeY=True, **kwargs)
    elif model == 'svr':
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, clsfr=LinearSVR(C=C), **kwargs)
    elif model == 'svc':
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, clsfr=LinearSVC(C=C), labelizeY=True, **kwargs)
    elif isinstance(model,sklearn.linear_model) or isinstance(model,sklearn.svm):
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, clsfr=model, labelizeY=True, **kwargs)
    elif model=='linearsklearn':
        clsfr = LinearSklearn(tau=tau, offset=offset, evtlabs=evtlabs, **kwargs)
    else:
        raise NotImplementedError("don't  know this model: {}".format(model))

    # fit the model
    if cv:
        res = clsfr.cv_fit(X, Y, cv=cv)
        Fy = res['estimator']
    else:
        print("Warning! overfitting...")
        clsfr.fit(X,Y)
        Fy = clsfr.predict(X, Y, dedup0=True)

    # assess model performance
    score=clsfr.audc_score(Fy)
    print(clsfr)
    print("score={}".format(score))
    (dc) = decodingCurveSupervised(Fy, priorsigma=(clsfr.sigma0_,clsfr.priorweight), softmaxscale=clsfr.softmaxscale_)

    return score,dc,Fy,clsfr

def analyse_datasets(dataset:str, model:str='cca', dataset_args:dict=None, loader_args:dict=None, preprocess_args:dict=None, clsfr_args:dict=None):
    """analyse a set of datasets (multiple subject) and generate a summary decoding plot.

    Args:
        dataset ([str]): the name of the dataset to load
        model (str, optional): The type of model to fit. Defaults to 'cca'.
        dataset_args ([dict], optional): additional arguments for get_dataset. Defaults to None.
        loader_args ([dict], optional): additional arguments for the dataset loader. Defaults to None.
        clsfr_args ([dict], optional): additional aguments for the model_fitter. Defaults to None.
    """    
    if dataset_args is None: dataset_args = dict()
    if loader_args is None: loader_args = dict()
    if clsfr_args is None: clsfr_args = dict()
    loader, filenames, dataroot = get_dataset(dataset,**dataset_args)
    scores=[]
    decoding_curves=[]
    nout=[]
    for i, fi in enumerate(filenames):
        print("{}) {}".format(i, fi))
        #try:
        if 1:
            X, Y, coords = loader(fi, **loader_args)
            if preprocess_args is not None:
                X, Y, coords = preprocess(X, Y, coords, **preprocess_args)
            score, decoding_curve, _, _ = analyse_dataset(X, Y, coords, model, **clsfr_args)
            nout.append(Y.shape[-1] if Y.ndim<=3 else Y.shape[-2])
            scores.append(score)
            decoding_curves.append(decoding_curve)
            del X, Y
            gc.collect()
        #except Exception as ex:
        #    print("Error: {}\nSKIPPED".format(ex))
    avescore=sum(scores)/len(scores)
    avenout=sum(nout)/len(nout)
    print("\n--------\n\n Ave-score={}\n".format(avescore))
    # extract averaged decoding curve info
    int_len, prob_err, prob_err_est, se, st = flatten_decoding_curves(decoding_curves)
    print("Ave-DC\n{}\n".format(print_decoding_curve(np.mean(int_len,0),np.mean(prob_err,0),np.mean(prob_err_est,0),np.mean(se,0),np.mean(st,0))))
    plot_decoding_curve(int_len,prob_err)
    plt.suptitle("{} ({}) AUDC={:3.2f}(n={} ncls={})\nloader={}\nclsfr={}({})".format(dataset,dataset_args,avescore,len(scores),avenout-1,loader_args,model,clsfr_args))
    plt.savefig("{}_decoding_curve.png".format(dataset))
    plt.show()

def flatten_decoding_curves(decoding_curves):
    ''' take list of (potentially variable length) decoding curves and make into a single array '''
    il = np.zeros((len(decoding_curves),decoding_curves[0][0].size))
    pe = np.zeros(il.shape)
    pee = np.zeros(il.shape)
    se = np.zeros(il.shape)
    st = np.zeros(il.shape)
    # TODO [] : insert according to the actual int-len
    for di,dc in enumerate(decoding_curves):
        il_di = dc[0]
        ll = min(il.shape[1],il_di.size)
        il[di,:ll] = dc[0][:ll]
        pe[di,:ll] = dc[1][:ll]
        pee[di,:ll] = dc[2][:ll]
        se[di,:ll] = dc[3][:ll]
        st[di,:ll] = dc[4][:ll] 
    return il,pe,pee,se,st

def debug_test_dataset(X, Y, coords=None, tau_ms=300, fs=None, offset_ms=0, evtlabs=('re', 'fe'), rank=1, model='cca', preprocess_args=None, clsfr_args=dict(), **kwargs):
    fs = coords[1]['fs'] if coords is not None else fs
    tau = int(fs*tau_ms/1000)
    offset=int(offset_ms*fs/1000)    
    times = np.arange(offset,tau+offset)/fs
    
    if coords is not None:
        print("X({}){}".format([c['name'] for c in coords], X.shape))
    else:
        print("X={}".format(X.shape))
    print("Y={}".format(Y.shape))
    print("fs={}".format(fs))

    if preprocess_args is not None:
        X, Y, coords = preprocess(X, Y, coords, **preprocess_args)

    ch_names = coords[2]['coords'] if coords is not None else None
    ch_pos = None
    if coords is not None and 'pos2d' in coords[2]:
        ch_pos = coords[2]['pos2d']
    elif not ch_names is None and len(ch_names) > 0:
        from mindaffectBCI.decoder.readCapInf import getPosInfo
        cnames, xy, xyz, iseeg =getPosInfo(ch_names)
        ch_pos=xy
    if ch_pos is not None:
        print('ch_pos={}'.format(ch_pos.shape))

    # visualize the dataset
    from mindaffectBCI.decoder.stim2event import stim2event
    from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, plot_erp, plot_summary_statistics, idOutliers
    import matplotlib.pyplot as plt

    print("Plot X+Y")
    trli=min(3,X.shape[0]-1)
    plt.figure(10); plt.clf()
    plt.subplot(211);
    plt.imshow(X[trli,:,:].T,aspect='auto');plt.colorbar();plt.title('X');plt.xlabel('time (samp)');plt.legend();
    plt.subplot(212);
    if Y.ndim == 3:
        plt.imshow(Y[trli, :, :].T, aspect='auto')
        plt.xlabel('time (samp)')
        plt.ylabel('target')
    else:
        plt.plot(Y[trli, :, 0, :])
    plt.title('Y')
    plt.show()

    print("Plot summary stats")
    if Y.ndim == 4: # already transformed
        Yevt = Y
    else: # convert to event
        Yevt = stim2event(Y, axis=-2, evtypes=evtlabs)
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Yevt[..., 0:1, :], tau=tau, offset=offset)
    plt.figure(11); plt.clf()
    plot_summary_statistics(Cxx, Cxy, Cyy, evtlabs, times, ch_names)
    plt.show()

    print('Plot global spectral properties')
    plot_grand_average_spectrum(X,axis=-2,fs=fs, ch_names=ch_names)
    plt.show()

    print("Plot ERP")
    plt.figure(12);plt.clf()
    plot_erp(Cxy, ch_names=ch_names, evtlabs=evtlabs, times=times)
    plt.suptitle("ERP")
    plt.show()
    
    # fit the model
    # override with direct keyword arguments
    clsfr_args['evtlabs']=evtlabs
    clsfr_args['tau_ms']=tau_ms
    clsfr_args['fs']=fs
    clsfr_args['offset_ms']=offset_ms
    clsfr_args['rank']=rank
    score, res, Fy, clsfr = analyse_dataset(X,Y,coords,model,**clsfr_args,**kwargs)
    
    plt.figure(14)
    plot_decoding_curve(*res)

    print("Plot Model")
    plt.figure(15);plt.clf()
    #filter2pattern(clsfr.sigma_,factored2full(clsfr.W_,clsfr.R_))
    if hasattr(clsfr,'A_'):
        plot_erp(factored2full(clsfr.A_, clsfr.R_), ch_names=ch_names, evtlabs=evtlabs, times=times)
        plt.suptitle("fwd-model")
    else:
        plot_erp(factored2full(clsfr.W_, clsfr.R_), ch_names=ch_names, evtlabs=evtlabs, times=times)
        plt.suptitle("bwd-model")
    plt.show()


    if not clsfr.R_ is None:
        print("Plot Factored Model")
        plt.figure(18)
        plt.clf()
        clsfr.plot_model(fs=fs)
        plt.show()
    
    print("plot Fe")
    plt.figure(16);plt.clf()
    Fe = clsfr.transform(X)
    plot_Fe(Fe)
    plt.suptitle("Fe")
    plt.show()

    print("plot Fy")
    plt.figure(17);plt.clf()
    plot_Fy(Fy,cumsum=True)
    plt.suptitle("Fy")
    plt.show()

    from mindaffectBCI.decoder.normalizeOutputScores import normalizeOutputScores, plot_normalizedScores
    print("normalized Fy")
    plt.figure(20);plt.clf()
    # normalize every sample
    ssFy, scale_sFy, decisIdx, nEp, nY = normalizeOutputScores(Fy, minDecisLen=-1)
    plot_Fy(ssFy,cumsum=False)
    plt.suptitle("normalized_Fy")
    plt.show()

    plt.figure(21)
    plot_normalizedScores(Fy[4,:,:],ssFy[4,:,:],scale_sFy[4,:],decisIdx)

    return clsfr,res

def debug_test_single_dataset(dataset:str,filename:str=None,dataset_args=None, loader_args=None, *args,**kwargs):
    """run the debug_test_dataset for a single subject from dataset

    Args:
        dataset ([str]): the dataset to load with get_dataset from `datasets.py`
        filename ([str], optional): a specific filename regular expression to match to process. Defaults to None.

    Returns:
        clsfr [BaseSeq2seq]: the model fitted during the dataset testing
    """    
    if dataset_args is None: dataset_args=dict()
    if loader_args is None: loader_args=dict()
    l,fs,_=get_dataset(dataset,**dataset_args)
    if filename is not None:
        fs = [f for f in fs if re.search(filename,f)]
    X,Y,coords=l(fs[0],**loader_args)
    return debug_test_dataset(X,Y,coords,*args,**kwargs)


    plt.close('all')
def run_analysis():    
    #analyse_datasets("plos_one",loader_args=dict(fs_out=60,stopband=((0,3),(30,-1))),
    #                 model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=3))
    #"plos_one",loader_args=dict(fs_out=120,stopband=((0,3),(45,-1))),model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1)): ave-score:67
    #"plos_one",loader_args=dict(fs_out=60,stopband=((0,3),(25,-1))),model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1)): ave-score:67
    #"plos_one",loader_args=dict(fs_out=60,stopband=((0,3),(25,-1))),model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=3)): ave-score:67
    #"plos_one",loader_args=dict(fs_out=60,stopband=((0,3),(45,-1))),model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1)): ave-score:674
    #"plos_one",loader_args=dict(fs_out=60,stopband=((0,2),(25,-1))),model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1)): ave-score:61
    #"plos_one",tau_ms=350,evtlabs=('re','fe'),rank=1 : ave-score=72  -- should be 83!!!
    # C: slightly larger freq range helps. rank doesn't.

    #analyse_datasets("lowlands",loader_args=dict(fs_out=60,stopband=((0,5),(25,-1))),
    #                  model='cca',clsfr_args=dict(tau_ms=350,evtlabs=('re','fe')))#,badEpThresh=6))
    #"lowlands",clsfr_args=dict(tau_ms=550,evtlabs=('re','fe'),rank=1,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,5),(25,-1))): ave-score=56
    #"lowlands",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,5),(25,-1))): ave-score=56
    #"lowlands",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=3,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,5),(25,-1))): ave-score=51
    #"lowlands",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=10,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,3),(25,-1))): ave-score=53
    #"lowlands",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=10,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,5),(25,-1))): ave-score=42
    #"lowlands",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=10,badEpThresh=6,rcond=1e-6),loader_args=(stopband=((0,5),(25,-1))): ave-score=45
    #analyse_datasets("lowlands",loader_args=dict(passband=None,stopband=((0,5),(25,-1))),clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=1,badEpThresh=4)): ave-score=.47

    #analyse_datasets("lowlands", clsfr_args=dict(tau_ms=350, evtlabs=('re','fe'), rank=3), loader_args=dict(passband=(4, 25),stopband=None))
    #"lowlands", tau_ms=350, evtlabs=('re','fe'),rank=1,loader_args={'passband':(5,25)} : Ave-score=0.64
    #"lowlands", tau_ms=700, evtlabs=('re'), rank=1, loader_args={'passband':(5, 25)}): ave-scre=.50
    #"lowlands", tau_ms=350, evtlabs=('re','fe'), rank=3, loader_args={'passband':(5, 25)}): score=.65
    #"lowlands", tau_ms=350, evtlabs=('re','fe'), rank=3, loader_args={'passband':(3, 25)}): .49
    # C: 5-25, rank=3, re+fe ~300ms
    # Q: why results so much lower now?

    # N.B. ram limits the  tau size...
    # analyse_datasets("brainsonfire",
    #                 loader_args=dict(fs_out=30, subtriallen=10, stopband=((0,1),(12,-1))),
    #                 model='cca',clsfr_args=dict(tau_ms=600, offset_ms=-300, evtlabs=None, rank=20))
    #"brainsonfire",loader_args=dict(fs_out=30, subtriallen=10, stopband=((0,1),(12,-1))),model='cca',clsfr_args=dict(tau_ms=600, offset_ms=-300, evtlabs=None, rank=5)) : score=.46
    #"brainsonfire",loader_args=dict(fs_out=30, subtriallen=10, stopband=((0,1),(12,-1))),model='cca',clsfr_args=dict(tau_ms=600, offset_ms=-300, evtlabs=None, rank=10)) : score=.53

    #analyse_datasets("twofinger",
    #                 model='cca',clsfr_args=dict(tau_ms=600, offset_ms=-300, evtlabs=None, rank=5), 
    #                 loader_args=dict(fs_out=60, subtriallen=10, stopband=((0,1),(25,-1))))
    #"twofinger",'cca',clsfr_args=dict(tau_ms=600, offset_ms=-300, evtlabs=None, rank=5),loader_args=dict(fs_out=60, subtriallen=10, stopband=((0,1),(25,-1)))): ave-score=.78
    # "twofinger",tau_ms=600, offset_ms=-300, rank=5,subtriallen=10, stopband=((0,1),(25,-1)))): ave-score: .85
    # C: slight benefit from pre-movement data
    
    # Note: max tau=500 due to memory limitation
    #analyse_datasets("cocktail",
    #                 clsfr_args=dict(tau_ms=500, evtlabs=None, rank=5, rcond=1e-4, center=False),
    #                 loader_args=dict(fs_out=60, subtriallen=10, stopband=((0,1),(25,-1))))
    #analyse_datasets("cocktail",tau_ms=500,evtlabs=None,rank=4,loader_args={'fs_out':60, 'subtriallen':15,'passband':(5,25)}) : .78
    #analyse_datasets("cocktail",tau_ms=500,evtlabs=None,rank=4,loader_args={'fs_out':60, 'subtriallen':15,'passband':(1,25)}) : .765
    #analyse_datasets("cocktail",tau_ms=500,evtlabs=None,rank=4,loader_args={'fs_out':30, 'subtriallen':15,'passband':(1,25)}) : .765
    #analyse_datasets("cocktail",tau_ms=500,evtlabs=None,rank=4,loader_args={'fs_out':30, 'subtriallen':15,'passband':(1,12)}) : .77
    #analyse_datasets("cocktail",tau_ms=500,evtlabs=None,rank=8,loader_args={'fs_out':30, 'subtriallen':15,'passband':(1,12)}) : .818
    #analyse_datasets("cocktail",tau_ms=700,evtlabs=None,rank=8,loader_args={'fs_out':30, 'subtriallen':15,'passband':(1,12)}) : .826
    #analyse_datasets("cocktail",tau_ms=700,evtlabs=None,rank=16,loader_args={'fs_out':30, 'subtriallen':15,'passband':(1,12)}) : .854
    #analyse_datasets("cocktail",tau_ms=500, evtlabs=None, rank=15,fs_out=60, subtriallen=10, stopband=((0,1),(25,-1)) : ave-score:.80 (6-subtrials)
    # C: longer analysis window + higher rank is better.  Sample rate isn't too important

    #analyse_datasets("openBMI_ERP",clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=5),loader_args=dict(fs_out=30,stopband=((0,1),(12,-1)),offset_ms=(-500,1000)))
    # "openBMI_ERP",tau_ms=700,evtlabs=('re'),rank=1,loader_args=dict(offset_ms=(-500,1000) Ave-score=0.758
    # "openBMI_ERP",tau_ms=700,evtlabs=('re','ntre'),rank=1,loader_args={'offset_ms':(-500,1000)}) Ave-score=0.822
    # "openBMI_ERP",tau_ms=700,evtlabs=('re','ntre'),rank=5,loader_args={'offset_ms':(-500,1000)}) Ave-score=0.894
    #"openBMI_ERP",clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=5),loader_args=dict(offset_ms=(-500,1000))): Ave-score=0.894
    # C: large-window, tgt-vs-ntgt  + rank>1 : gives best fit?

    #analyse_datasets("openBMI_SSVEP",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=6),loader_args=dict(offset_ms=(-500,1000)))
    # "openBMI_SSVEP",tau_ms=700,evtlabs=('re'),rank=1,loader_args={'offset_ms':(-500,1000)} : score=.942
    # "openBMI_SSVEP",tau_ms=700,evtlabs=('re'),rank=6,loader_args={'offset_ms':(-500,1000)} : score=.947
    # "openBMI_SSVEP",tau_ms=700,evtlabs=('re','fe'),rank=1,loader_args={'offset_ms':(-500,1000)} : score= :.745
    # "openBMI_SSVEP",tau_ms=350,evtlabs=('re','fe'),rank=6,loader_args={'offset_ms':(-500,1000)} : score= .916
    #analyse_datasets("openBMI_SSVEP",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=6),loader_args=dict(offset_ms=(-500,1000))) : score=.917
    #analyse_datasets("openBMI_SSVEP",clsfr_args=dict(tau_ms=350,evtlabs=('re','fe'),rank=6),loader_args=dict(offset_ms=(-500,1000))) : score=.92
    # "openBMI",tau_ms=600,evtlabs=('re'),rank=1,loader_args={'offset_ms':(-500,1000)} : score==.940
    # C: large-window, re, rank>1 : gives best fit?
    
    #analyse_datasets("p300_prn",loader_args=dict(fs_out=30,stopband=((0,1),(25,-1)),subtriallen=10),
    #                 model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=5))
    #"p300_prn",model='cca',loader_args=dict(fs_out=30,stopband=((0,2),(12,-1)),subtriallen=10),clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=15)) : score=.43
    #"p300_prn",model='cca',loader_args=dict(fs_out=60,stopband=((0,2),(25,-1)),subtriallen=10),clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=15)) : score=.47

    #analyse_datasets("mTRF_audio", tau_ms=600, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(5, 25)})
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(.5, 15)}) : score=.86
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=2, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(.5, 15)}) : score=.85
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(5, 25)}) : score = .89
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(.5, 25)}) : score = .86
    #analyse_datasets("mTRF_audio", tau_ms=100, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(5, 25)}) : score= .85
    #analyse_datasets("mTRF_audio", tau_ms=20, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(5, 25)}) : score=.88
    #analyse_datasets("mTRF_audio", tau_ms=600, evtlabs=None, rank=5, loader_args={'regressor':'spectrogram', 'fs_out':64, 'passband':(5, 25)}) : score=.91
    
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'envelope', 'fs_out':64, 'passband':(.5, 15)}) : score=.77
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'envelope', 'fs_out':64, 'passband':(5, 25)}) : score=.77
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=5, loader_args={'regressor':'envelope', 'fs_out':128, 'passband':(5, 25)}) : score=.78
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=2, loader_args={'regressor':'envelope', 'fs_out':128, 'passband':(5, 25)}) : score=.76
    #analyse_datasets("mTRF_audio", tau_ms=300, evtlabs=None, rank=1, loader_args={'regressor':'envelope', 'fs_out':128, 'passband':(5, 25)}) : score=.69

    # C: spectrogram (over envelope), rank>3, 5-25Hz, short tau is sufficient ~ 100ms

    #analyse_datasets("tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),
    #                 model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=5))
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=10) : ave-score:51
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=5) : ave-score:54
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=3) : ave-score:54
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(12,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=3) : ave-score:52
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(12,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','ntre'),rank=10) : ave-score:49
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','anyre'),rank=5) : ave-score:54
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','anyre'),rank=10) : ave-score:50
    #"tactileP3",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re'),rank=5) : ave-score:44
    # C: above chance for 8/9, low rank~3, slow response
    
    #analyse_datasets("tactile_PatientStudy",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),
    #                 model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','anyre'),rank=5))
    #"tactile_PatientStudy",loader_args=dict(fs_out=60,stopband=((0,1),(25,-1))),model='cca',clsfr_args=dict(tau_ms=700,evtlabs=('re','anyre'),rank=5) : ave-score:44

    #analyse_datasets("ninapro_db2",loader_args=dict(stopband=((0,15), (45,55), (95,105), (250,-1)), fs_out=60, nvirt=20, whiten=True, rectify=True, log=True, plot=False, filterbank=None, zscore_y=True),
    #                 model='cca',clsfr_args=dict(tau_ms=40,evtlabs=None,rank=6))
    #"ninapro_db2",loader_args=dict(subtrllen=10, stopband=((0,15), (45,55), (95,105), (250,-1)), fs_out=60, nvirt=40, whiten=True, rectify=True, log=True, plot=False, filterbank=None, zscore_y=True),model='cca',clsfr_args=dict(tau_ms=40,evtlabs=None,rank=20)): ave-score=65 (but dont' believe it)
    #"ninapro_db2",loader_args=dict(subtrllen=10, stopband=((0,15), (45,55), (95,105), (250,-1)), fs_out=60, nvirt=40, whiten=True, rectify=True, log=True, plot=False, filterbank=None, zscore_y=True),model='ridge',clsfr_args=dict(tau_ms=40,evtlabs=None,rank=20)): ave-score=26 (but dont' believe it)

    #analyse_datasets("openBMI_MI",clsfr_args=dict(tau_ms=350,evtlabs=None,rank=6),loader_args=dict(offset_ms=(-500,1000)))
    pass


if __name__=="__main__":

    #debug_test_single_dataset('p300_prn',dataset_args=dict(label='rc_5_flash'),
    #              loader_args=dict(fs_out=32,stopband=((0,1),(12,-1)),subtriallen=None),
    #              model='cca',tau_ms=750,evtlabs=('re','anyre'),rank=3,reg=.02)

    from offline.load_mindaffectBCI  import load_mindaffectBCI
    savefile = None
    #savefile = '/Users/Developer/Desktop/UtopiaMessages_2007191_1558_brainproducts.log'
    #savefile = '/Users/Developer/Desktop/mindaffectBCI_200720_2152_testing_python.txt'
    #savefile = '/Users/Developer/Desktop/mindaffectBCI_200720_2147_testing_octave.txt'
    #savefile = '/Users/Developer/Desktop/mindaffectBCI_200720_2128_master.txt'
    #savefile = '/Users/Developer/Desktop/mindaffectBCI_200720_2116_testing_gdx.txt'
    #savefile = "/Users/Developer/Downloads/mindaffectBCI_200717_1625_mark_testing_octave_gdx.txt"
    if savefile is None:
        # default to last log file if not given
        import glob
        import os
        files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
        savefile = max(files, key=os.path.getctime)
    X, Y, coords = load_mindaffectBCI(savefile, stopband=((45,65),(0,3),(25,-1)), fs_out=60)
    debug_test_dataset(X, Y, coords, tau_ms=400, evtlabs=('re','fe'), rank=1, model='cca')
    #debug_test_dataset(X, Y, coords, tau_ms=400, evtlabs=('re','fe'), rank=1, model='lr', ignore_unlabelled=True)

    #run_analysis()



import numpy as np
from mindaffectBCI.decoder.utils import window_axis

#@function
def scoreStimulus(X, W, R=None, b=None, offset=0, isepoched=None):
    '''
    Apply spatio-temporal (possibly factored) model to data 
    Inputs:
      X = (nTrl x nSamp x d) raw response for the current stimulus event
               d=#electrodes tau=#response-samples  nEpoch=#stimulus events to process
         OR
          (nTrl x nEpoch x tau x d) pre-sliced raw data
      W = (nM x nfilt x d) spatial filters for each output
      R = (nM x nfilt x nE x tau) responses for each stimulus event for each output
     OR
      W = (nM x nE x tau x d) spatio-temporal filter per event type and model
      R = None
      b = (nE,1) bias for each stimulus type
      offset = 0 (1,1) offset in X for applying W
    Outputs:
      Fe= (nM x nTrl x nEpoch/nSamp x nE) similarity score for each input epoch for each output
    Copyright (c) MindAffect B.V. 2018
    '''
    tau = W.shape[-2] if R is None else R.shape[-1] # est response length
    if isepoched is None: # guess epoched state from the shapes...
        isepoched = X.ndim > 3
    if isepoched:
        Fe = scoreStimulusEpoch(X, W, R, b)
    else:
        Fe = scoreStimulusCont(X, W, R, b, offset)
    return Fe

def scoreStimulusEpoch(X, W, R=None, b=None):
    '''
    Apply spatio-temporal (possibly factored) model to epoched data 
      X = (nTrl x nEpoch x tau x d) pre-sliced raw data
      W = (nM x nfilt x d) spatial filters for each output
      R = (nM x nfilt x nE x tau) responses for each stimulus event for each output
     OR
      W = (nM x nE x tau x d) spatio-temporal filter per event type and model
      R = None
      b = (nE,1) offset for each stimulus type
    Outputs:
      Fe= (nM x nTrl x nEpoch/nSamp x nE) similarity score for each input epoch for each output
    '''
    if R is not None:
        Fe = scoreStimulusEpoch_factored(X,W,R,b)
    else:
        Fe = scoreStimulusEpoch_full(X,W,b)
    return Fe


def scoreStimulusEpoch_factored(X, W, R, b=None):
    '''
    Apply factored spatio-temporal model to epoched data 
      X = (nTrl x nEpoch x tau x d) pre-sliced raw data
      W = (nM x nfilt x d) spatial filters for each output
      R = (nM x nfilt x nE x tau) responses for each stimulus event for each output
      b = (nE,1) offset for each stimulus type
    Outputs:
      Fe= (nM x nTrl x nEpoch x nE) similarity score for each input epoch for each output
    '''

    # ensure all inputs have the  right  shape
    if X.ndim < 4:
        X = X.reshape((1,)*(4-X.ndim)+X.shape)
    if W.ndim < 3:
        W = W.reshape(((1,)*(3-W.ndim))+W.shape)
    if R.ndim < 4:
        R = R.reshape(((1,)*(4-R.ndim))+R.shape)

    # apply the factored model, complex product to optimize the path
    Fe = np.einsum("Mkd,TEtd,Mket->MTEe", W, X, R, optimize='optimal') 
    #Fe = np.einsum("Mkd,TEtd->TEMkt",W,X) # manual factored
    #Fe = np.einsum("TEMkt,Mket->MTEe",Fe,R)
    if not b is None: # include the bias, for each stimulus type
        Fe = Fe + b
    return Fe


def scoreStimulusEpoch_full(X, W, b=None):
    '''
    Apply full spatio-temporal model to epoched data 
      X = (nTrl x nEpoch x tau x d) pre-sliced raw data
      W = (nM x nE x tau x d) spatio-temporal filter per event type and model
      b = (nE,1) offset for each stimulus type
    Outputs:
      Fe= (nM x nTrl x nEpoch/nSamp x nE) similarity score for each input epoch for each output
    '''

    # ensure inputs have the  right  shape
    if X.ndim < 4:
        X = X.reshape((1,)*(4-X.ndim)+X.shape)
    if W.ndim < 4:
        W = W.reshape((1,)*(4-W.ndim)+W.shape)

    # apply the model
    Fe = np.einsum("TEtd, metd->mTEe", X, W, optimize='optimal')
    return Fe

def factored2full(W, R):
    ''' convert a factored spatio-temporal model to a full model
    Inputs:
       W (nM,rank,d) spatial filter set (BWD model)
       R (nM,rank,e,tau) temporal filter set (FWD model)
    Output:
       W (nM,e,tau,d) spatio-temporal filter (BWD model) '''
    if R is not None:
        if W.ndim < 3:
            W = W.reshape(((1,)*(3-W.ndim))+W.shape)
        if R.ndim < 4:
            R = R.reshape(((1,)*(4-R.ndim))+R.shape)
        # get to single spatio-temporal filter
        W = np.einsum("mfd, mfet->metd", W, R)
    return W

def scoreStimulusCont(X, W, R=None, b=None, offset=0):
    '''
    Apply spatio-tempoal (possibly factored) model to raw (non epoched) data
    '''
    W = factored2full(W, R) # (metd)
    tau = W.shape[-2]
    if X.shape[-2] > tau:
        Xe = window_axis(X, winsz=tau, axis=-2) # (nTrl, nSamp-tau, tau, d)
        Fe = scoreStimulusEpoch(Xe, W, None, b) # (nM, nTrl, nSamp-tau, nE)
    else: # X isn't big enough to apply... => zero score
        Fe = np.zeros((W.shape[0],X.shape[0],1,W.shape[1]),dtype=X.dtype)
    
    # zero pad to keep the output size
    # TODO[]: pad with the *actual* values...
    tmp = np.zeros(Fe.shape[:-2]+(X.shape[-2], Fe.shape[-1]),dtype=X.dtype)
    # padd after the result = accumulate information *backwards* in time to
    # start of the window -> Fe causes X
    # also shift by the desired offset to compensate for shifted windows
    tmp[..., -offset:Fe.shape[-2]-offset, :] = Fe
    Fe = tmp

    return Fe

    
def plot_Fe(Fe):
    import matplotlib.pyplot as plt
    '''plot the stimulus score function'''
    #print("Fe={}".format(Fe.shape))
    if Fe.ndim > 3:
        if Fe.shape[0]>1 :
            print('Warning stripping model dimension!')
        Fe = Fe[0,...]
    plt.clf()
    nPlts=min(25,Fe.shape[0])
    if Fe.shape[0]/2 > nPlts:
        tis = np.linspace(0,Fe.shape[0]/2-1,nPlts,dtype=int)
    else:
        tis = np.arange(0,nPlts,dtype=int)
    ncols = int(np.ceil(np.sqrt(nPlts)))
    nrows = int(np.ceil(nPlts/ncols))
    axploti = ncols*(nrows-1)
    ax = plt.subplot(nrows,ncols,axploti+1)
    for ci,ti in enumerate(tis):
        # make the axis
        if ci==axploti: # common axis plot
            pl = ax
        else: # normal plot
            pl = plt.subplot(nrows,ncols,ci+1,sharex=ax, sharey=ax) # share limits
            plt.tick_params(labelbottom=False,labelleft=False) # no labels        
        pl.plot(Fe[ti,:,:])
        pl.set_title("{}".format(ti))
    pl.legend(range(Fe.shape[-1]-1))
    plt.suptitle('Fe')

#@function
def testcase():
    #   X = (nTrl, nEp, tau, d) [d x tau x nEpoch x nTrl ] raw response for the current stimulus event
    #            d=#electrodes tau=#response-samples  nEpoch=#stimulus events to process
    #    w = (nM, nfilt, d) spatial filters for each output
    #    r = (nM, nfilt, nE, tau) responses for each stimulus event for each output
    nE = 2
    d = 8
    tau = 10
    nSamp = 300
    nTrl = 30
    nfilt = 1
    nM = 20
    X = np.random.randn(nTrl, nSamp, d)
    Xe = window_axis(X, winsz = tau, axis = -2)
    W = np.random.randn(nM, nfilt, d)
    R = np.random.randn(nM, nfilt, nE, tau)
    Fe = scoreStimulus(Xe, W, R)
    print("Xe={} -> Fe={}".format(Xe.shape, Fe.shape))
    
    Wf = factored2full(W,R)
    Fef = scoreStimulus(Xe,Wf)
    print("Wf={} -> Fef={}".format(Wf.shape,Fef.shape))

    print("Fe-Fef={}".format(np.max(np.abs(Fe-Fef).ravel())))

    F = scoreStimulusCont(X, W, R)
    print("X={} -> F={}".format(X.shape, F.shape))

if __name__=="__main__":
    testcase()



#!/usr/bin/env python3
import numpy as np
from mindaffectBCI.decoder.UtopiaDataInterface import UtopiaDataInterface, stim2eventfilt, butterfilt_and_downsample

def sigViewer(ui: UtopiaDataInterface=None, hostname=None, timeout_ms:float=np.inf, center=True, timerange:int=5, nstimulus_lines:int=1, ndata_lines:int=-1, datastep:int=20, stimstep:int=1):
    ''' simple sig-viewer using the ring-buffer for testing '''
    import matplotlib.pyplot as plt

    if ui is None:
        data_preprocessor = butterfilt_and_downsample(order=6, stopband=((0,3),(25,-1)), fs_out=60)#999)
        #data_preprocessor = butterfilt_and_downsample(order=6, stopband='butter_stopband((0, 5), (25, -1))_fs200.pk', fs_out=60)
        ui=UtopiaDataInterface(data_preprocessor=data_preprocessor, send_signalquality=False)#, sample2timestamp='none')
        ui.connect(hostname)

    ui.update()

    # initialize the plot window
    fig = plt.figure(1)
    fig.clear()    
    
    # initialize the plot window
    plt.clf()
    plt.axes((.05,.25,.9,.7))
    plt.title("EEG")
    idx=slice(-int(ui.fs*timerange),None)
    data = ui.data_ringbuffer[idx, :]
    # vertical gap between EEG lines
    data_linestep = np.arange(data[:, :ndata_lines].shape[-1]) * datastep
    # compute the x-point,  i.e. in time. Use these as ring-buffer may not be full yet
    xdata = np.arange(-data.shape[0], 0) / ui.fs
    data_lines = plt.plot(xdata, data[:, :ndata_lines] + data_linestep[np.newaxis,:])
    plt.ylim([-datastep, data[:,:ndata_lines].shape[1] * datastep])
    plt.grid(True)

    plt.axes((.05,.05,.9,.1))
    plt.title("Stimulus")
    stimulus = ui.stimulus_ringbuffer[idx, :]
    # vertical gap between stimulus lines
    stimulus_linestep = np.arange(nstimulus_lines) * stimstep
    stimulus_lines = plt.plot((xdata[0],xdata[-1]),np.zeros((2,nstimulus_lines))+stimulus_linestep)
    plt.ylim([-stimstep, nstimulus_lines* stimstep])
    plt.grid(True)
    #plt.autoscale(axis='x')

    if False:
        f2=plt.figure(2)
        timing_line = plt.plot(1000*(xdata-np.arange(-xdata.shape[0], 0) / ui.fs))[0]
        plt.xlabel('sample')
        plt.ylabel('error w.r.t. sampling @ {}hz (ts - ts*fs) (ms)'.format(ui.fs))
        plt.ylim(-200,200)
        plt.grid()
        f2.show()

    # start the render loop
    fig.show()
    t0 = ui.getTimeStamp()
    while ui.getTimeStamp() < t0+timeout_ms:

        # exit if figure is closed
        if not plt.fignum_exists(1):
            quit()
        
        # Update the records, at least 100ms wait..
        ui.update(timeout_ms=100, mintime_ms=95)

        # Update the EEG stream
        idx = slice(-int(ui.fs*timerange),None) # final 5s data
        data = ui.data_ringbuffer[idx, :]
        if center:
            # only the non-timestamp channels
            data[:,:-1] = data[:,:-1] - np.mean(data[-int(data.shape[0]*.25):,:-1],axis=0)
        xdata = ( data[idx,-1] - data[-1,-1] ) / 1000 # time-position in seconds, relative to last
        for li, ln in enumerate(data_lines):
            ln.set_xdata(xdata)
            ln.set_ydata(data[idx, li] + data_linestep[li])

        if False:
            plt.figure(2)
            timing_line.set_xdata(xdata)
            timing_line.set_ydata(1000*(xdata-np.arange(-xdata.shape[0], 0) / ui.fs))
            f2.canvas.draw()
            f2.canvas.flush_events()

        # Update the Stimulus Stream
        # TODO[]: search backwards instead of assuming the stimulus rate...
        stimulus = ui.stimulus_ringbuffer[idx,:]
        # BODGE: pad with 2 extra 0 events, if last stimulus before last data
        if stimulus.shape[0]<2 or stimulus[-1,-1] < data[-1,-1]:
            pad = np.zeros((2,stimulus.shape[1]),dtype=stimulus.dtype)
            pad[0,-1] = stimulus[-1,-1]+1 if stimulus.shape[0]>0 else data[-1,-1]-1000*timerange
            pad[1,-1] = data[-1,-1]
            stimulus = np.append(stimulus,pad,axis=0)
        xdata = ( stimulus[idx,-1] - stimulus[-1,-1] ) / 1000
        for li, ln in enumerate(stimulus_lines):
            ln.set_xdata(xdata)
            ln.set_ydata(stimulus[idx, li] + stimulus_linestep[li])

        # re-draw and sleep for the next one
        fig.canvas.draw()
        fig.canvas.flush_events()

if __name__=='__main__':
    import sys
    print("Args: {}".format(sys.argv))
    if len(sys.argv) > 1:
        hostname=sys.argv[1]
    else:
        hostname=None

    try:
        matplotlib.use('Qt4Cairo')
    except:
        pass

    sigViewer()



#!/bin/env python3
#

# Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from utopiaclient import *
import subprocess
class Utopia2Harmony:    
    """Example class for a utopia OUTPUT module.  Connects to the utopia server
    and listens for output which exceed it's probability threshold before
    then printing them and using NEWTARGET to indicated the output has taken
    place
    """
    
    def __init__(self,outputPressThreshold=.1,outputReleaseThreshold=None,myObjectIDs=None):
        self.VERB=0
        self.outputPressThreshold=outputPressThreshold
        if outputReleaseThreshold is not None :
            self.outputReleaseThreshold=outputPressThreshold*2
        else:
            self.outputReleaseThreshold=outputReleaseThreshold
        self.myObjectIDs=myObjectIDs # this is the set of object IDs we generate output for
        self.client = UtopiaClient()

    def connect(self,host=UtopiaClient.DEFAULTHOST,port=UtopiaClient.DEFAULTPORT,timeout_ms=30000):
        self.client.autoconnect(host,port,timeout_ms)
        self.client.initClockAlign()

    def perrModeOutput(self,msg):
        """process a perr-message generating appropriate output.  
           To avoid 'key-bounce' we use a press-release semantics, where the output is 'Pressed' 
           so the output is generated when the Perr < outputPressThreshold, then further
           output is inhibited until Perr > outputReleaseThreshold."""
        if self.myObjectIDs is None or msg.Yest in self.myObjectIDs : # one of "our" outputs
            if msg.Perr < self.outputPressThreshold and not self.outputActivated : # low enough error, not already activated
                self.doOutputHarmony(msg.Yest) # call function to actually make the output happen
                # TODO [] : send a selection message rather than NewTarget to indicate *what* was selected
                self.client.sendMessage(NewTarget(self.client.getTimeStamp())) # notify we've made selection
                self.outputActivated=True # update state to output activated
            elif msg.Perr > self.outputReleaseThreshold and self.outputActivated : # high-enough error, already activated
                self.outputActivated=False # relase output -> allow a new selection

    def selectionModeOutput(self,msg):
        """ Process selection message to generate output.  
            Basically generate output if the messages objectID is one of the ones
            we are tasked with generating output for"""
        if self.myObjectIDs is None or msg.objID in self.myObjectIDs : # it's one of "our" outputs
            self.doOutputHarmony(msg.objID) # call function to make output happen

    def run(self,timeout_ms=3000):
        """mainloop of utopia-to-output mapping
        runs an infinite loop, waiting for new messages from utopia, filtering out 
        those mesages which contain an output prediction (i.e. PREDICTEDTARGETPROB message)
        and if the output prediction is sufficiently confident forwarding this to the output
        device and sending a NEWTARGET to the recogniser to indicate the output was sent
        """
        if not self.client.isConnected :
            self.connect()
        print("Waiting for messages")
        self.outputActivated=False
        while True:
            newmsgs = self.client.getNewMessages(timeout_ms)
            for msg in newmsgs:
                if self.VERB>1 : print("Got message " + str(msg) + " <- server")
                if msg.msgID == PredictedTargetProb.msgID and not self.outputPressThreshold is None :
                    # Perr message and in Perr based output mode
                    self.perrModeOutput(msg)
                    
                elif msg.msgID == Selection.msgID and self.outputPressThreshold is None :
                    # Selection message in in Selection based output mode
                    self.selectionModeOutput(msg)
                if self.VERB>0 : print('.')


            
    def doOutputHarmony(self,objID):
        """This function is run when objID has sufficiently low error to mean that 
        and output should be generated for this objID. 
        N.B. Override/Replace this function with your specific output method."""
        print("Generated output for Target %d"%(objID))
        #Set the ip address of the harmony hub
        ip_address = "192.168.1.31"
        # set the device id
        id_Telenet = "63961272"
        id_Apple_TV = "63944910"
        # check for selection messages and send the related infrared/bluetooth signal
        if objID ==96 : 	  
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "PowerToggle"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==97 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "VolumeUp"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==98 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "VolumeDown"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==99 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "ChannelUp"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==100 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "ChannelDown"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==101 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Telenet, "--command", "Mute"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==104 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "DirectionUp"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==105 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "DirectionRight"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==106 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "DirectionDown"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==107 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "DirectionLeft"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==108 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "select"]);
            print("Target %d is good enough!"%(objID))
        elif objID ==109 : 
            rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "menu"]);
            print("Target %d is good enough!"%(objID))
        else :
           # rtn = subprocess.call(["aioharmony","--harmony_ip",ip_address,"send_command","--device_id", id_Apple_TV, "--command", "menu"]);
            print("an Unprogrammed button (ID = %d) is chosen!"%(objID))


''' simple driver for testing '''
if __name__ == "__main__":
    import sys
    host = UtopiaClient.DEFAULTHOST
    port = UtopiaClient.DEFAULTPORT
    
    if len(sys.argv)>1:
        hostname = sys.argv[1]
        tmp=[hostname.split(":")]
        if len(tmp)>1:
            hostname=tmp[0];
            port    =tmp[1]

    outputthreshold=.1
    if len(sys.argv)>2:
        try:
            outputthreshold = float(sys.argv[2])
        except:
            print('argument should be a valid floating point threshold'%sys.argv[2])
            
    u2harmony = Utopia2Harmony(None)
    u2harmony.connect(host,port)
    u2harmony.run()



import os
import numpy as np
from scipy.io import loadmat
from mindaffectBCI.decoder.utils import butter_sosfilt

def load_brainstream(datadir, sessdir=None, sessfn=None, fs_out=60, ifs=None, fr=None, passband=None, stopband=((45,65),(0,5),(25,-1)), verb=0, ch_names=None):

    # load the data file
    Xfn = datadir
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)
    data = loadmat(Xfn)
    if 'v' in data.keys(): # silly mat struct stuff..
        data = data['v'] # [d x samp x trl ]
    #print("data.keys={}".format(data.keys()))

    if 'X' in data: # all in 1 file, raw plos_one format
        X  = data['X'] # (d x samp x trl)
        cb = data['codes'] if 'codes' in data else data['V'] # (samples x nY)
        lab = data['y'] # (trl)

    elif 'data' in data: # pre-converted data
        X    = data['data']     # [ d x samp_seq x nTrl ]
        cb   = data['codebooks'] # [ samp_code x nY ]
        lab  = data['labels']   # [ nTrl x 1] # index into codes array

    else: # extract info from other files...

        X = data
        trainmode = 'train' in fn
        # general config = codebook info
        cfgfn = os.path.join(datadir, sessdir, 'cfg.mat')
        try:
            cfg = loadmat(cfgfn)
            stim = cfg['stimulation']
            fr = stim['rate']*stim['uprate']
            if trainmode:
                subset = stim['trainsubset']
                layout = stim['trainlayout']
                cb = stim['U'] # [ samp_code x nY ]
            else:
                subset = stim['testsubset']
                layout = stim['testlayout']
                cb = stim['V'] # [ samp_code x nY ]
            cb = cb[:, trainlayout]
            cb = cb[:, trainlayout] # [ samp_code x nY ]
        except:
            raise ValueError('Couldnt load the configuration file!')

        # extract the header to get the sample rate
        hdrfn = os.path.join(datadir, sessdir, 'hdr.mat')
        try:
            hdr=loadmat(hdrfn)
            if 'v' in hdr.keys():
                hdr = hdr['v']
                ifs = hdr['Fs'][0]
        except:
            print('Warning: Couldnt load the header file');

        # load per-trial labels
        if trainmode:
            labfn = os.path.join(datadir, sessdir, 'train'+'labels.mat')
        else:
            labfn = os.path.join(datadir, sessdir, 'test'+'labels.mat')
        lab = loadmat(labfn)
        if 'v' in lab.keys(): lab = lab['v'] # [ trl ]

    # get the sample rate,  and downsample rate
    if ifs is None: # get the sample rate info
        if 'fs' in data:
            ifs = data['fs'][0][0]
        elif 'fSample' in data:
            ifs = data['fSample'][0][0]
        else:
            ifs = 180
    if fr is None:
        if 'fr' in data:
            fr = data['fr'][0][0]
        else:
            fr = 60
    fs = ifs
    if fs_out is None:
        fs_out = ifs
        
    X  = X.astype("float32") # [ ch x samp x trl ]:float - raw eeg
    X  = np.moveaxis(X, (0, 1, 2), (2, 1, 0)) # (nTrl, nSamp, d)
    X  = np.ascontiguousarray(X) # ensure memory efficient layout
    lab = lab.astype("uint8").ravel() # [ trl ]:int - index into codebook
    cb = cb.astype("bool") # [ samp x nY ]:bool -- codebook

    # convert lab+code into Y + samp-times,  where 1st row is always the true label
    cb = cb[np.mod(np.arange(0, X.shape[1]), cb.shape[0]), :] # loop cb up to X size [ samp x nY]
    Y  = np.zeros((len(lab), X.shape[1], cb.shape[1]+1), dtype=bool) # (nTr, nSamp, nY) [nY x samp x trl]
    for ti, l in enumerate(lab):
        Y[ti, :, 0]  = cb[:, l-1] # copy in true label
        Y[ti, :, 1:] = cb # copy in other outputs

    # preprocess -> spectral filter
    if stopband is not None or passband is not None:
        # BODGE: pre-center X
        X = X - X[...,0:1,:]
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X, stopband, fs, passband=passband)
    
    # preprocess -> downsample
    resamprate = round(2*fs/fs_out)/2 # round to nearest .5
    if resamprate > 1:
        if 1 or verb > 0:
            print("resample by {}: {}->{}Hz".format(resamprate, fs, fs/resamprate))
        idx = np.arange(0,X.shape[1],resamprate).astype(np.int)
        X = X[:, idx, :] # decimate X (trl, samp, d)
        Y = Y[:, idx, :] # decimate Y (trl, samp, y)
        fs = fs/resamprate

    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1]) *1000/fs, \
                 'fs':fs}
    coords[2] = {'name':'channel', 'coords':ch_names}
    
    return (X, Y, coords)

def testcase():
    import sys

    # plos_one
    #datadir = '/home/jadref/removable/SD Card/data/bci/own_experiments/noisetagging_v3'
    #sessdir = 's01'
    #sessfn = 'traindata.mat'

    # lowlands
    datadir = '/home/jadref/removable/SD Card/data/bci/own_experiments/lowlands'
    sessdir = ''
    sessfn = 'LL_ENG_39_20170820_tr_train_1.mat' # should perform 100%

    # command-line, for testing
    if len(sys.argv) > 1:
        datadir = sys.argv[1]
    if len(sys.argv) > 2:
        sessdir = sys.argv[2]
    if len(sys.argv) > 3:
        fn = sys.argv[3]

    from load_brainstream import load_brainstream
    X, Y, coords = load_brainstream(datadir, sessdir, sessfn, fs_out=60, stopband=((0,5.5),(24,-1)))
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']
    
    from model_fitting import MultiCCA
    from decodingCurveSupervised import decodingCurveSupervised
    cca = MultiCCA(tau=18)
    cca.fit(X, Y)
    print('cca = {}'.format(cca))
    Fy = cca.predict(X, Y, dedup0=True)
    print("Fy={}".format(Fy.shape))
    (_) = decodingCurveSupervised(Fy)

    # test w.r.t. matlab
    from scipy.io import loadmat
    from utils import window_axis
    import numpy as np
    Xe = window_axis(X, axis=-2, winsz=18)
    Ye = window_axis(Y, axis=-2, winsz=1)
    
    matdata=loadmat('/home/jadref/'+sessfn)
    Xm = np.moveaxis(matdata['X'], (0, 1, 2), (2, 1, 0)) # (t,s,d)
    Ym = np.moveaxis(matdata['Y'], (0, 1, 2), (2, 1, 0)) # (t,s,y)
    stem = np.moveaxis(matdata['stimTimes_samp'], (0, 1), (1, 0)) # (t,e)

    cca = MultiCCA(tau=18)
    cca.fit(Xm, Ym)
    cca.score(Xm, Ym)
    (res) = decodingCurveSupervised(cca.predict(Xm,Ym))
    
    print("X-Xm={}".format(np.max(np.abs(X-Xm).ravel())))
    print("Y-Ym={}".format(np.max(np.abs(Y[:, :, 0]-Ym[:, :, 0]).ravel())))

    Xem = np.moveaxis(matdata['Xe'], (0, 1, 2, 3), (3, 2,1,0))# (t,e,tau,d)
    Yem = np.moveaxis(matdata['Ye'], (0, 1, 2), (2, 1, 0)) # (t,e, y)

    # off by 1 in the sliced version, w.r.t. Matlab
    print("Xe-Xem={}".format(np.max(np.abs(Xe[:,1:Xem.shape[1]+1,:,:]-Xem).ravel())))
    
    print("Ye-Yem={}".format(np.max(np.abs(Ye[:,1:Yem.shape[1]+1, 0, 0]-Yem[:, :, 0]).ravel())))
    
    
    cca = MultiCCA(tau=Xem.shape[-2])
    cca.fit(Xem, Yem, stem)
    print('cca = {}'.format(cca))
    Fy = cca.predict(Xem, Yem)
    print("Fy={}".format(Fy.shape))
    (res)=decodingCurveSupervised(Fy);

    # run in stages
    from updateSummaryStatistics import updateSummaryStatistics
    from stim2event import stim2event
    Yeem=np.moveaxis(matdata['Yee'], (0, 1, 2, 3), (3, 2, 1, 0))
    Yeme=stim2event(Yem, ['re', 'fe'], -2) # py convert to brain events
    print("Yeem-Yeme={}".format(np.max(np.abs(Yeem-Yeme).ravel())))
    
    Cxx,Cxy,Cyy=updateSummaryStatistics(Xem, Yeem[:,:,0:1,:], stem)

    Cxxm=matdata['Cxx']
    Cxym=np.moveaxis(matdata['Cxy'],(0,1,2),(2,1,0))
    Cyym=np.moveaxis(matdata['Cyy'],(0,1,2,3),(3,2,1,0))

    print('Cxx-Cxxm={}'.format(np.max(np.abs(Cxx-Cxxm).ravel())))
    print('Cxy-Cxym={}'.format(np.max(np.abs(Cxy-Cxym).ravel())))
    print('Cyy-Cyym={}'.format(np.max(np.abs(Cyy-Cyym).ravel())))
    
    import matplotlib.pyplot as plt
    plt.clf();plt.plot(X[1,:,0]);plt.plot(Xm[0,:,0])
    plt.clf();plt.plot(Y_true.ravel());plt.plot(Ym_true.ravel());

if __name__ == "__main__":
    testcase()



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mindaffectBCI.noisetag import Noisetag, sumstats


import pyglet
# define a simple 2-squares drawing function
def draw_squares(col1,col2):
    # draw square 1: @100,190 , width=100, height=100
    x=100; y=190; w=100; h=100;
    pyglet.graphics.draw(4,pyglet.gl.GL_QUADS,
                         ('v2f',(x,y,x+w,y,x+w,y+h,x,y+h)),
             ('c3f',(col1)*4))
    # draw square 2: @440,100
    x=640-100-100
    pyglet.graphics.draw(4,pyglet.gl.GL_QUADS,
                         ('v2f',(x,y,x+w,y,x+w,y+h,x,y+h)),
             ('c3f',(col2)*4))    

# dictionary mapping from stimulus-state to colors
state2color={0:(.2,.2,.2), # off=grey
             1:(1,1,1),    # on=white
             2:(0,1,0),    # cue=green
             3:(0,0,1)}    # feedback=blue
def draw(dt):
    '''draw the display with colors from noisetag'''
    # send info on the *previous* stimulus state, with the recorded vsync time (if available)
    fliptime = window.lastfliptime if window.lastfliptime else nt.getTimeStamp()
    nt.sendStimulusState(timestamp=fliptime)
    # update and get the new stimulus state to display
    try : 
        nt.updateStimulusState()
        stimulus_state,target_state,objIDs,sendEvents=nt.getStimulusState()
    except StopIteration :
        pyglet.app.exit() # terminate app when noisetag is done
        return

    # draw the display with the instructed colors
    if stimulus_state : 
        draw_squares(state2color[stimulus_state[0]],
                     state2color[stimulus_state[1]])

    # some textual logging of what's happening
    if target_state is not None and target_state>=0:
        print("*" if target_state>0 else '.',end='',flush=True)
    else:
        print('.',end='',flush=True)

# WARNING: Hacking ahead!!!
# override window's flip method to record the exact *time* the
# flip happended
import types
def timedflip(self):
    '''pseudo method type which records the timestamp for window flips'''
    type(self).flip(self) # call the 'real' flip method...
    self.lastfliptime=nt.getTimeStamp()

# define a trival selection handler
def selectionHandler(objID):
    print("Selected: %d"%(objID))    


# Initialize the noise-tagging connection
nt = Noisetag()
nt.connect(timeout_ms=5000)
nt.addSelectionHandler(selectionHandler)
# tell the noisetag framework to run a full : calibrate->prediction sequence
nt.setnumActiveObjIDs(2)
nt.startExpt(nCal=2,nPred=10,duration=4)

# Initialize the drawing window
# make a default window, with fixed size for simplicty, and vsync for timing
config = pyglet.gl.Config(double_buffer=True)
window = pyglet.window.Window(width=640,height=480, vsync=True, config=config)

# Setup the flip-time recording for this window
window.flip = types.MethodType(timedflip,window)
window.lastfliptime=None
    
# run the pyglet main loop
pyglet.clock.schedule(draw)
pyglet.app.run()



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import pyglet
import sys
import time
import math

#import os
#scriptpath=os.path.dirname(os.path.realpath(__file__))
#import sys
#sys.path.append(os.path.join(scriptpath,'../../'))

from mindaffectBCI.noisetag import Noisetag, sumstats

class Quad:
    '''object to hold a single graphics quad'''
    def __init__(self,objID,vertices,color,theta):
        self.objID=objID
        self.vertices=vertices
        self.color=color
        self.theta=theta
    def draw(self,maskcolor=None,alpha=.5):
        '''draw this quad, optionally with given color'''
        ''' given color overrides'''
        if maskcolor :
            # use alpha to blend color and maskcolor
            color = tuple(maskcolor[i]*alpha + self.color[i]*(1-alpha) for i in range(len(self.color)))
        else : 
            color=self.color
        # draw colored segement
        pyglet.graphics.draw(4,pyglet.gl.GL_QUADS,
                            ('v2f',self.vertices),
                            ('c3f',(color)*4))

class DrawWindow(pyglet.window.Window):

    def __init__(self,noisetag,*args,**kwargs):
        super(DrawWindow,self).__init__(*args,**kwargs)
        self.frame=0
        self.segments=[]
        self.objIDs=[]
        self.noisetag=None
        self.lastfliptime=0
        self.flipstats=sumstats()
        self.fliplogtime=self.lastfliptime
        self.noisetag=noisetag
        
    def init_pie(self,nsegements=12,iradius=1/6,oradius=1/3,objIDs=None):
        self.objIDs = objIDs
        if self.objIDs is None :
            self.objIDs = list(range(1,nsegements+1))
        if nsegements is None :
            nsegements = len(self.objIDs)
        winw,winh=self.get_size()
        cent   =(winw/2,winh/2)
        iradius=min([winw,winh])*iradius
        oradius=min([winw,winh])*oradius
        segwidth= math.pi*2 / nsegements
        # center of the segement
        thetas=[ i*segwidth for i in range(nsegements) ]

        # TODO [] : make a pyglet BATCH...
        self.segements=[]
        for i,theta in enumerate(thetas):
            # inner-top
            it = (cent[0]+iradius*math.sin(theta),
                  cent[1]+iradius*math.cos(theta))
            # outer-top
            ot = (cent[0]+oradius*math.sin(theta),
                  cent[1]+oradius*math.cos(theta))
            # outer-bot
            ob = (cent[0]+oradius*math.sin(theta+segwidth),
                  cent[1]+oradius*math.cos(theta+segwidth))
            # inner-bot
            ib = (cent[0]+iradius*math.sin(theta+segwidth),
                  cent[1]+iradius*math.cos(theta+segwidth))

            # color
            colsegwidth=2.0*math.pi/3.0
            if   0*colsegwidth <=theta and theta<= 1*colsegwidth : 
                alpha = (theta-0*colsegwidth)/(colsegwidth)
                bg    = ((1-alpha), alpha,       0) # r-g
            elif 1*colsegwidth <=theta and theta<= 2*colsegwidth :
                alpha = (theta-1*colsegwidth)/(colsegwidth)
                bg    = (0,        (1-alpha),   alpha) # g-b
            elif 2*colsegwidth <=theta and theta<= 3*colsegwidth : 
                alpha = (theta-2*colsegwidth)/(colsegwidth)
                bg    = (alpha,       0,     (1-alpha)) # b-r
            
            #print('theta %g  alpha %g = (%g,%g,%g)'%(theta,alpha,*bg))
            # each segment is a single graphics quad
            self.segements.append(Quad(self.objIDs[i],it+ot+ob+ib,bg,theta))

    def setNoisetag(self,noisetag):
        self.noisetag=noisetag

    # mapping between bci-stimulus-states and on-screen colors
    state2color={0:(.2,.2,.2), # off=grey
                 1:(1,1,1),    # on=white
                 2:(1,1,1)}    # cue=flash
    def draw_pie(self):
        self.frame=self.frame+1
        # get the bci-stimulus state
        stimulus_state=None
        if self.noisetag :
            self.noisetag.sendStimulusState(timestamp=self.lastfliptime)
            self.noisetag.updateStimulusState()
            stimulus_state,target_state,objIDs,sendEvents=self.noisetag.getStimulusState()
        # do nothing if no bci-stimulus
        if stimulus_state is None :
            stimulus_state = [0]*len(self.segements)

        if target_state is not None and target_state>0:
            print("*" if target_state>0 else '.',end='',flush=True)

        self.clear()        
        # modify the blend strength if in cue/feedback mode
        alpha=.6
        if 1 in stimulus_state :
            alpha=.7 # medium background in flicker mode
        elif 3 in stimulus_state :
            alpha=.9 # dark background if feedback mode
        elif 2 in stimulus_state : 
            alpha=.6 # medium background in cue-mode
        # draw the segements with correct bci-stim mask blending
        for i,seg in enumerate(self.segements):
            # color depends on the requested stimulus state
            try:
                bg=self.state2color[stimulus_state[i]]
            except KeyError :
                bg=None # normal color
            seg.draw(maskcolor=bg,alpha=alpha)            

    def update(self,dt):
        self.draw_pie()
        
    def flip(self):
        # override window's flip method to send stimulus state as close in
        # time as possible to when the screen re-freshed
        super().flip()
        olft=self.lastfliptime
        self.lastfliptime=self.noisetag.getTimeStamp()
        self.flipstats.addpoint(self.lastfliptime-olft)
        if self.lastfliptime > self.fliplogtime :
            self.fliplogtime=self.lastfliptime+10000    
            print("\nFlipTimes:"+str(self.flipstats))
            print("Hist:\n"+self.flipstats.hist())
            
def selectionHandler(objID):
    '''function to map selection to action'''    
    print('Sel: %d'%(objID))
    # if is obj we are responsible for
    if objID in window.objIDs :
        # get the matching segment
        segidx = window.objIDs.index(objID)
        seg    = window.segements[segidx]
        # set the light to this color, by matching the hue.
        if hue_bridge : 
            try : 
                lights=hue_bridge.lights()
                for k,v in lights.items():
                    if v['state']['reachable'] :
                        hue_bridge.lights[k]('state',bri=128,hue=int(65535*seg.theta/(2*math.pi)))
            except :
                print("couldnt talk to the hue!") 


if __name__=='__main__':
    # N.B. init the noisetag first so asks for decoder IP
    noisetag=Noisetag()
    # auto connect to the decoder
    noisetag.connect()

    config = pyglet.gl.Config(double_buffer=True)
    window = DrawWindow(noisetag,width=1024,height=768,vsync=True,config=config)
    window.init_pie(nsegements=10)

    # set the current noise-tagging state
    noisetag.setActiveObjIDs(window.objIDs)
    noisetag.startExpt(nCal=10,nPred=20,
                       cueduration=2,calduration=4,predduration=20,
                       feedbackduration=4)

    # Initialize the connection to the hue light
    try :
        from qhue import Bridge
        bridgeip="192.168.253.100"
        username="AlGbD1GTQHxwDVK0j3tKkpdwBUv8Cijtvbkokzxk"
        hue_bridge = Bridge(bridgeip, username)
        # register selection handler
        noisetag.addSelectionHandler(selectionHandler)
    except ImportError :
        import warnings
        warnings.warn("You need to install the qhue module to connect to your hue!")
    
    # run the mainloop
    pyglet.clock.schedule(window.update)
    pyglet.app.run()



#!/usr/bin/env python3
import socket
import sys
import time
import re
import struct

def ip_is_local(ip_string):
    """
    Uses a regex to determine if the input ip is on a local network. Returns a boolean. 
    It's safe here, but never use a regex for IP verification if from a potentially dangerous source.
    """
    combined_regex = "(^10\.)|(^172\.1[6-9]\.)|(^172\.2[0-9]\.)|(^172\.3[0-1]\.)|(^192\.168\.)"
    return re.match(combined_regex, ip_string) is not None # is not None is just a sneaky way of converting to a boolean

def get_ip_address( NICname ):
    import fcntl
    import struct
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        s.fileno(),
        0x8915,  # SIOCGIFADDR
        struct.pack('256s', NICname[:15].encode("UTF-8"))
    )[20:24])

def nic_info():
    """
    Return a list with tuples containing NIC and IPv4 address
    """
    nic = []
    for ix in socket.if_nameindex():
        name = ix[1]
        ip = get_ip_address( name )

        nic.append( (name, ip) )
    return nic    

def get_all_ips():
    try:
        ips = [ i[1] for i in nic_info() ]
    except:
        # fall back on getaddrinfo
        ips = [ x[4][0] for x in socket.getaddrinfo(socket.gethostname(), 80) ]
    return ips

def get_local_ip():
    # socket.getaddrinfo returns a bunch of info, so we just get the IPs it returns with this list comprehension.
    local_ips = [ i for i in get_all_ips() if ip_is_local(i) ]
    # prefer one with internet access..
    print("all local ips: {}".format(local_ips))
    local_ip = local_ips[0] if len(local_ips) > 0 else '127.0.0.1'
    return local_ip

def get_remote_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(.5)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()
    except socketerror:
        ip = None
    return ip

class ssdpDiscover :
    ssdpv4group = ("239.255.255.250", 1900)
    ssdpv6group= ("FF08::C", 1900)
    ssdpgroup = None

    msearchTemplate = "\r\n".join([
        'M-SEARCH * HTTP/1.1',
        'HOST: {0}:{1}',
        'MAN: "ssdp:discover"',
        'ST: {st}', 'MX: {mx}', '', ''])
    def __init__(self,servicetype="ssdp:all",discoverytimeout=3):
        self.servicetype=servicetype
        self.queryt=None
        self.sock=None

    def makeDiscoveryMessage(self,ssdpgroup,servicetype,timeout):
        msearchMessage = self.msearchTemplate.format(*ssdpgroup, st=servicetype, mx=timeout)
        if sys.version_info[0] == 3:
            msearchMessage = msearchMessage.encode("utf-8")
        return msearchMessage
    
    def initSocket(self, v6=False):
        # make the UDP socket to the multicast group with timeout
        if v6 and socket.has_ipv6:
            IPPROTO_IPV6 = socket.IPPROTO_IPV6 if hasattr(socket,'IPPROTO_V6') else 41
            self.ssdpgroup = self.ssdpv6group
            self.sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                self.sock.setsockopt(IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 5)
                self.sock.setsockopt(IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 1)
                self.sock.setsockopt(IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, 0)
            except:
                print("couldn't set socket options\n")
            self.sock.bind(('',self.ssdpgroup[1])) # bind port

            # request multicast group membership
            mreq = struct.pack("16s15s".encode('utf-8'), 
                                socket.inet_pton(socket.AF_INET6, self.ssdpgroup[0]), 
                                (chr(0) * 16).encode('utf-8'))
            self.sock.setsockopt(IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, mreq)

            if_idx=0
            self.sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, struct.pack("I",if_idx))

        else:
            self.ssdpgroup = self.ssdpv4group
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 5)
            self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1)

            self.sock.bind(('',self.ssdpgroup[1])) # bind port

            # request membership
            group = socket.inet_aton(self.ssdpgroup[0])

            # Set multicast interface -- only on unix!
            if_idx=0
            #ifi = socket.if_nametoindex(ifn)
            #ifis = struct.pack("I", ifi)
            self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, struct.pack("I",if_idx))

            #local_ip = get_remote_ip()        
            #if local_ip is None:
            #   local_ip=get_local_ip()
            #print("Trying local ip: {}".format(local_ip))
            #mreq = group + socket.inet_aton(local_ip)

            mreq = struct.pack('4sl', group, socket.INADDR_ANY)

            # add to multi-cast group to get responses
            self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        
        
    def discover(self,timeout=.001,querytimeout=5,v6=False):
        '''auto-discover the utopia-hub using ssdp discover messages,
           timeout is time to wait for a response.  query timeout is time between re-sending the ssdp-discovery query message.'''
        # make and send the discovery message
        if self.sock is None:
            try : 
                self.initSocket(v6)
            except socket.error :
                print("Couldnt init socket!")
                return ()

        # re-send the ssdp query
        if self.queryt is None or self.queryt+querytimeout < time.time():
            # build discovery message for this service/group
            self.msearchMessage=self.makeDiscoveryMessage(self.ssdpgroup, self.servicetype, querytimeout)

            print("Sending query message:\n%s"%(self.msearchMessage))
            try:
                self.sock.sendto(self.msearchMessage, self.ssdpgroup)
                self.queryt = time.time()
            except socket.error as ex: # abort if send fails
                print("send error")
                print(ex)
                return ()

        # wait for the specified time for a response
        t0 = time.time()
        ttg = timeout
        responses = []
        while time.time() < t0 + timeout:
            location, rsp = self.wait_msearch_response(ttg)
            if location is not None:
                responses.append(location)
            ttg = t0 + timeout - time.time()

        # wait for responses to our query
        return responses

    def wait_msearch_response(self,timeout):
        self.sock.settimeout(timeout)
        print("Waiting responses")
        try:
            rsp,addr=self.sock.recvfrom(8192)
        except socket.timeout:
            print("Socket timeout")
            return (None,None)
        except socket.error as ex :
            print("Socket error" + str(ex)) 
            return (None,None)

        if rsp == self.msearchMessage:
            print("Response was query message!")
            return (None,None)


        # got a response, parse it...           
        rsp=rsp.decode('utf-8')
        print("Got response from : %s\n%s\n"%(addr,rsp))

        # guard for recieving our own query message!
        #if "M-SEARCH" in rsp:

        # does the response contain servertype, if so then we match
        location=None
        if self.servicetype is None or self.servicetype in rsp :
            print("Response matches server type: %s"%(self.servicetype))
            # use the message source address as default location
            location=addr[0] if hasattr(addr,'__iter__') else addr 
            # extract the location or IP from the message
            for line in rsp.split("\r\n"): # cut into lines
                tmp=line.split(":",1) # cut line into key/val
                # is self the key we care about -> then record the value
                if len(tmp)>1 and tmp[0].lower()=="LOCATION".lower() :
                    location=tmp[1].strip()
                    # strip http:// xxxx /stuff
                    if location.startswith("http://"):
                        location=location[7:] # strip http
                    if '/' in location :
                        location=location[:location.index('/')] # strip desc.xml
                    print("Got location: {}".format(location))
                    break # done with self response
            # add to the list of possible servers
            print("Loc added to response list: {}".format(location))
        return (location,rsp)


def ipscanDiscover(port:int, ip:str=None, timeout=.5):
    ''' scan for service by trying all 255 possible final ip-addresses'''
    if ip is None:
        ip = get_remote_ip()
    if ip is None:
        #    local_ip=get_local_ip()
        ips = get_all_ips()
        # prefer non local-host IP if there is one.
        nonlocalip = [ i for i in ips if not i.startswith('127.') ]
        print("nonlocalhost={}".format(nonlocalip))        
        ip = nonlocalip[0] if nonlocalip else ips[0]
        
    ipprefix = ".".join(ip.split('.')[:3])
    hosts = []
    for postfix in range(255):
        ip = "{}.{:d}".format(ipprefix,postfix)
        try:
            print("Trying: {}:{}".format(ip,port))
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)
            sock.connect((ip,port))            
            print("Connected")
            hosts.append(ip)
            sock.close()
        except socket.error as ex:
            pass
    return hosts


def testIncrementalScanning():
    disc=ssdpDiscover("utopia/1.1")
    while True:
        d0 = time.time()
        resp = disc.discover(timeout=.001,v6=False)
        dt = time.time()-d0
        print("Discover time: " + str(dt))
        if len(resp)==0 :
            print("No new responses")
        else :
            print(resp)
            break
        # sleep to represent doing other work..
        time.sleep(1)

def discoverOrIPscan(port:int = 8400, timeout_ms:int = 15000):
    disc=ssdpDiscover("utopia/1.1")
    hosts = disc.discover(timeout=timeout_ms/1000.0,v6=False)
    
    if hosts is None or len(hosts) == 0:
        hosts=ipscanDiscover(8400)
    print("got hosts: {}".format(hosts))
    return hosts

if __name__=="__main__":
    try:
        print( nic_info() )
    except:
        pass
    
    discoverOrIPscan()


# Pre-processing:
#  filter = 1-12Hz -> stop band 0-1, 12-inf
#  downsample = 30hz
from mindaffectBCI.decoder.UtopiaDataInterface import UtopiaDataInterface, butterfilt_and_downsample
stopband = ((0, 1), (12, -1))
fs_out = 30
ppfn = butterfilt_and_downsample(order=6, stopband=stopband, fs_out=fs_out)
ui = UtopiaDataInterface(data_preprocessor=ppfn) 

# Classifier:
#   * response length 700ms (as the p300 is from 300-600ms)
tau_ms = 700
#   * start of target stimulus vs. start of any stimuls
#       -> 'rising-edge' and 'non-target rising edge'
evtlabs = ('re', 'ntre')
#   * rank-3 decomposition, as we tend to get multiple component, e.g. perceptual, P3a, P3b
rank = 3
#  CCA classifier
from mindaffectBCI.decoder.model_fitting import MultiCCA
clsfr = MultiCCA(tau=int(fs_out*tau_ms/1000), rank=rank, evtlabs=evtlabs)

from mindaffectBCI.decoder import decoder
decoder.mainloop(ui=ui, clsfr=clsfr, calplots=True)